@@ -14,9 +14,8 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-#include "got_compat.h"
-
 #include <sys/queue.h>
+#include <sys/tree.h>
 #include <sys/types.h>
 
 #include <ctype.h>
@@ -25,11 +24,14 @@
 #include <errno.h>
 #include <locale.h>
 #include <inttypes.h>
+#include <sha1.h>
+#include <sha2.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <signal.h>
 #include <string.h>
 #include <unistd.h>
+#include <util.h>
 
 #include "got_version.h"
 #include "got_error.h"
@@ -38,6 +40,8 @@
 #include "got_cancel.h"
 #include "got_repository.h"
 #include "got_repository_admin.h"
+#include "got_repository_dump.h"
+#include "got_repository_load.h"
 #include "got_gotconfig.h"
 #include "got_path.h"
 #include "got_privsep.h"
@@ -85,6 +89,8 @@ __dead static void	usage_pack(void);
 __dead static void	usage_indexpack(void);
 __dead static void	usage_listpack(void);
 __dead static void	usage_cleanup(void);
+__dead static void	usage_dump(void);
+__dead static void	usage_load(void);
 
 static const struct got_error*		cmd_init(int, char *[]);
 static const struct got_error*		cmd_info(int, char *[]);
@@ -92,6 +98,8 @@ static const struct got_error*		cmd_pack(int, char *[]);
 static const struct got_error*		cmd_indexpack(int, char *[]);
 static const struct got_error*		cmd_listpack(int, char *[]);
 static const struct got_error*		cmd_cleanup(int, char *[]);
+static const struct got_error*		cmd_dump(int, char *[]);
+static const struct got_error*		cmd_load(int, char *[]);
 
 static const struct gotadmin_cmd gotadmin_commands[] = {
 	{ "init",	cmd_init,	usage_init,	"" },
@@ -100,6 +108,8 @@ static const struct gotadmin_cmd gotadmin_commands[] = {
 	{ "indexpack",	cmd_indexpack,	usage_indexpack,"ix" },
 	{ "listpack",	cmd_listpack,	usage_listpack,	"ls" },
 	{ "cleanup",	cmd_cleanup,	usage_cleanup,	"cl" },
+	{ "dump",	cmd_dump,	usage_dump,	"" },
+	{ "load",	cmd_load,	usage_load,	"" },
 };
 
 static void
@@ -248,7 +258,7 @@ get_repo_path(char **repo_path)
 	if (cwd == NULL)
 		return got_error_from_errno("getcwd");
 
-	err = got_worktree_open(&worktree, cwd);
+	err = got_worktree_open(&worktree, cwd, NULL);
 	if (err) {
 		if (err->code != GOT_ERR_NOT_WORKTREE)
 			goto done;
@@ -271,7 +281,8 @@ done:
 __dead static void
 usage_init(void)
 {
-	fprintf(stderr, "usage: %s init [-b branch] repository-path\n",
+	fprintf(stderr, "usage: %s init [-A hashing-algorithm] [-b branch]"
+	    " repository-path\n",
 	    getprogname());
 	exit(1);
 }
@@ -282,6 +293,7 @@ cmd_init(int argc, char *argv[])
 	const struct got_error *error = NULL;
 	const char *head_name = NULL;
 	char *repo_path = NULL;
+	enum got_hash_algorithm algo = GOT_HASH_SHA1;
 	int ch;
 
 #ifndef PROFILE
@@ -289,8 +301,17 @@ cmd_init(int argc, char *argv[])
 		err(1, "pledge");
 #endif
 
-	while ((ch = getopt(argc, argv, "b:")) != -1) {
+	while ((ch = getopt(argc, argv, "A:b:")) != -1) {
 		switch (ch) {
+		case 'A':
+			if (!strcmp(optarg, "sha1"))
+				algo = GOT_HASH_SHA1;
+			else if (!strcmp(optarg, "sha256"))
+				algo = GOT_HASH_SHA256;
+			else
+				return got_error_path(optarg,
+				    GOT_ERR_OBJECT_FORMAT);
+			break;
 		case 'b':
 			head_name = optarg;
 			break;
@@ -321,7 +342,7 @@ cmd_init(int argc, char *argv[])
 	if (error)
 		goto done;
 
-	error = got_repo_init(repo_path, head_name);
+	error = got_repo_init(repo_path, head_name, algo);
 done:
 	free(repo_path);
 	return error;
@@ -385,6 +406,9 @@ cmd_info(int argc, char *argv[])
 		goto done;
 
 	printf("repository: %s\n", got_repo_get_path_git_dir(repo));
+	printf("hashing algorithm: %s\n",
+	    got_repo_get_object_format(repo) == GOT_HASH_SHA1 ? "sha1"
+	    : "sha256");
 
 	gotconfig = got_repo_get_gotconfig(repo);
 	if (gotconfig) {
@@ -458,6 +482,7 @@ usage_pack(void)
 }
 
 struct got_pack_progress_arg {
+	FILE *out;
 	char last_scaled_size[FMT_SCALED_STRSIZE];
 	int last_ncolored;
 	int last_nfound;
@@ -474,20 +499,20 @@ struct got_pack_progress_arg {
 };
 
 static void
-print_load_info(int print_colored, int print_found, int print_trees,
+print_load_info(FILE *out, int print_colored, int print_found, int print_trees,
     int ncolored, int nfound, int ntrees)
 {
 	if (print_colored) {
-		printf("%d commit%s colored", ncolored,
+		fprintf(out, "%d commit%s colored", ncolored,
 		    ncolored == 1 ? "" : "s");
 	}
 	if (print_found) {
-		printf("%s%d object%s found",
+		fprintf(out, "%s%d object%s found",
 		    ncolored > 0 ? "; " : "",
 		    nfound, nfound == 1 ? "" : "s");
 	}
 	if (print_trees) {
-		printf("; %d tree%s scanned", ntrees,
+		fprintf(out, "; %d tree%s scanned", ntrees,
 		    ntrees == 1 ? "" : "s");
 	}
 }
@@ -495,7 +520,7 @@ print_load_info(int print_colored, int print_found, int print_trees,
 static const struct got_error *
 pack_progress(void *arg, int ncolored, int nfound, int ntrees,
     off_t packfile_size, int ncommits, int nobj_total, int nobj_deltify,
-    int nobj_written)
+    int nobj_written, int pack_done)
 {
 	struct got_pack_progress_arg *a = arg;
 	char scaled_size[FMT_SCALED_STRSIZE];
@@ -527,16 +552,16 @@ pack_progress(void *arg, int ncolored, int nfound, int ntrees,
 
 	if ((print_colored || print_found || print_trees) &&
 	    !a->loading_done) {
-		printf("\r");
-		print_load_info(print_colored, print_found, print_trees,
-		    ncolored, nfound, ntrees);
+		fprintf(a->out, "\r");
+		print_load_info(a->out, print_colored, print_found,
+		    print_trees, ncolored, nfound, ntrees);
 		a->printed_something = 1;
-		fflush(stdout);
+		fflush(a->out);
 		return NULL;
 	} else if (!a->loading_done) {
-		printf("\r");
-		print_load_info(1, 1, 1, ncolored, nfound, ntrees);
-		printf("\n");
+		fprintf(a->out, "\r");
+		print_load_info(a->out, 1, 1, 1, ncolored, nfound, ntrees);
+		fprintf(a->out, "\n");
 		a->loading_done = 1;
 	}
 
@@ -584,29 +609,30 @@ pack_progress(void *arg, int ncolored, int nfound, int ntrees,
 	}
 
 	if (print_searching || print_total || print_deltify || print_written)
-		printf("\r");
+		fprintf(a->out, "\r");
 	if (print_searching)
-		printf("packing %d reference%s", ncommits,
+		fprintf(a->out, "packing %d reference%s", ncommits,
 		    ncommits == 1 ? "" : "s");
 	if (print_total)
-		printf("; %d object%s", nobj_total,
+		fprintf(a->out, "; %d object%s", nobj_total,
 		    nobj_total == 1 ? "" : "s");
 	if (print_deltify)
-		printf("; deltify: %d%%", p_deltify);
+		fprintf(a->out, "; deltify: %d%%", p_deltify);
 	if (print_written)
-		printf("; writing pack: %*s %d%%", FMT_SCALED_STRSIZE - 2,
-		    scaled_size, p_written);
-	if (print_searching || print_total || print_deltify ||
-	    print_written) {
+		fprintf(a->out, "; writing pack: %*s %d%%",
+		    FMT_SCALED_STRSIZE - 2, scaled_size, p_written);
+	if (print_searching || print_total || print_deltify || print_written) {
 		a->printed_something = 1;
-		fflush(stdout);
+		fflush(a->out);
 	}
+	if (pack_done)
+		fprintf(a->out, "\n");
 	return NULL;
 }
 
 static const struct got_error *
 pack_index_progress(void *arg, off_t packfile_size, int nobj_total,
-    int nobj_indexed, int nobj_loose, int nobj_resolved)
+    int nobj_indexed, int nobj_loose, int nobj_resolved, int indexing_done)
 {
 	struct got_pack_progress_arg *a = arg;
 	char scaled_size[FMT_SCALED_STRSIZE];
@@ -653,7 +679,9 @@ pack_index_progress(void *arg, off_t packfile_size, int nobj_total,
 		printf("; indexing %d%%", p_indexed);
 	if (print_resolved)
 		printf("; resolving deltas %d%%", p_resolved);
-	if (print_size || print_indexed || print_resolved)
+	if (indexing_done)
+		printf("\n");
+	if (print_size || print_indexed || print_resolved || indexing_done)
 		fflush(stdout);
 
 	return NULL;
@@ -694,7 +722,7 @@ cmd_pack(int argc, char *argv[])
 	struct got_repository *repo = NULL;
 	int ch, i, loose_obj_only = 1, force_refdelta = 0, verbosity = 0;
 	struct got_object_id *pack_hash = NULL;
-	char *id_str = NULL;
+	char *id_str = NULL, *idxpath = NULL;
 	struct got_pack_progress_arg ppa;
 	FILE *packfile = NULL;
 	struct got_pathlist_head exclude_args;
@@ -704,7 +732,7 @@ cmd_pack(int argc, char *argv[])
 	struct got_reflist_entry *re, *new;
 	int *pack_fds = NULL;
 
-	TAILQ_INIT(&exclude_args);
+	RB_INIT(&exclude_args);
 	TAILQ_INIT(&exclude_refs);
 	TAILQ_INIT(&include_refs);
 
@@ -734,7 +762,7 @@ cmd_pack(int argc, char *argv[])
 			break;
 		case 'x':
 			got_path_strip_trailing_slashes(optarg);
-			error = got_pathlist_append(&exclude_args,
+			error = got_pathlist_insert(NULL, &exclude_args,
 			    optarg, NULL);
 			if (error)
 				return error;
@@ -764,7 +792,7 @@ cmd_pack(int argc, char *argv[])
 	if (error)
 		goto done;
 
-	TAILQ_FOREACH(pe, &exclude_args, entry) {
+	RB_FOREACH(pe, got_pathlist_head, &exclude_args) {
 		const char *refname = pe->path;
 		error = add_ref(&new, &exclude_refs, refname, repo);
 		if (error)
@@ -798,6 +826,7 @@ cmd_pack(int argc, char *argv[])
 	}
 
 	memset(&ppa, 0, sizeof(ppa));
+	ppa.out = stdout;
 	ppa.last_scaled_size[0] = '\0';
 	ppa.last_p_indexed = -1;
 	ppa.last_p_resolved = -1;
@@ -818,7 +847,7 @@ cmd_pack(int argc, char *argv[])
 	if (verbosity >= 0)
 		printf("\nWrote %s.pack\n", id_str);
 
-	error = got_repo_index_pack(packfile, pack_hash, repo,
+	error = got_repo_index_pack(&idxpath, packfile, pack_hash, repo,
 	    pack_index_progress, &ppa, check_cancelled, NULL);
 	if (error)
 		goto done;
@@ -839,6 +868,7 @@ done:
 	free(id_str);
 	free(pack_hash);
 	free(repo_path);
+	free(idxpath);
 	return error;
 }
 
@@ -857,7 +887,7 @@ cmd_indexpack(int argc, char *argv[])
 	struct got_repository *repo = NULL;
 	int ch;
 	struct got_object_id *pack_hash = NULL;
-	char *packfile_path = NULL;
+	char *packfile_path = NULL, *idxpath = NULL;
 	char *id_str = NULL;
 	struct got_pack_progress_arg ppa;
 	FILE *packfile = NULL;
@@ -899,6 +929,7 @@ cmd_indexpack(int argc, char *argv[])
 		goto done;
 
 	memset(&ppa, 0, sizeof(ppa));
+	ppa.out = stdout;
 	ppa.last_scaled_size[0] = '\0';
 	ppa.last_p_indexed = -1;
 	ppa.last_p_resolved = -1;
@@ -912,7 +943,7 @@ cmd_indexpack(int argc, char *argv[])
 	if (error)
 		goto done;
 
-	error = got_repo_index_pack(packfile, pack_hash, repo,
+	error = got_repo_index_pack(&idxpath, packfile, pack_hash, repo,
 	    pack_index_progress, &ppa, check_cancelled, NULL);
 	if (error)
 		goto done;
@@ -928,6 +959,7 @@ done:
 	}
 	free(id_str);
 	free(pack_hash);
+	free(idxpath);
 	return error;
 }
 
@@ -1135,53 +1167,72 @@ struct got_cleanup_progress_arg {
 	int last_nloose;
 	int last_ncommits;
 	int last_npurged;
+	int last_nredundant;
 	int verbosity;
 	int printed_something;
 	int dry_run;
 };
 
 static const struct got_error *
-cleanup_progress(void *arg, int nloose, int ncommits, int npurged)
+cleanup_progress(void *arg, int ncommits, int nloose, int npurged,
+    int nredundant)
 {
 	struct got_cleanup_progress_arg *a = arg;
 	int print_loose = 0, print_commits = 0, print_purged = 0;
+	int print_redundant = 0;
 
-	if (a->last_nloose != nloose) {
-		print_loose = 1;
-		a->last_nloose = nloose;
-	}
 	if (a->last_ncommits != ncommits) {
-		print_loose = 1;
 		print_commits = 1;
 		a->last_ncommits = ncommits;
 	}
-	if (a->last_npurged != npurged) {
-		print_loose = 1;
+	if (a->last_nloose != nloose) {
 		print_commits = 1;
+		print_loose = 1;
+		a->last_nloose = nloose;
+	}
+	if (a->last_npurged != npurged) {
+		print_commits = 1;
+		print_loose = 1;
 		print_purged = 1;
 		a->last_npurged = npurged;
 	}
+	if (a->last_nredundant != nredundant) {
+		print_commits = 1;
+		print_loose = 1;
+		print_purged = 1;
+		print_redundant = 1;
+		a->last_nredundant = nredundant;
+	}
 
 	if (a->verbosity < 0)
 		return NULL;
 
-	if (print_loose || print_commits || print_purged)
+	if (print_loose || print_commits || print_purged || print_redundant)
 		printf("\r");
-	if (print_loose)
-		printf("%d loose object%s", nloose, nloose == 1 ? "" : "s");
 	if (print_commits)
-		printf("; %d commit%s scanned", ncommits,
+		printf("%d commit%s scanned", ncommits,
 		    ncommits == 1 ? "" : "s");
-	if (print_purged) {
+	if (print_loose)
+		printf("; %d loose object%s", nloose, nloose == 1 ? "" : "s");
+	if (print_purged || print_redundant) {
 		if (a->dry_run) {
-			printf("; %d object%s could be purged", npurged,
+			printf("; could purge %d object%s", npurged,
 			    npurged == 1 ? "" : "s");
 		} else {
-			printf("; %d object%s purged", npurged,
+			printf("; purged %d object%s", npurged,
 			    npurged == 1 ? "" : "s");
 		}
 	}
-	if (print_loose || print_commits || print_purged) {
+	if (print_redundant) {
+		if (a->dry_run) {
+			printf(", %d pack file%s", nredundant,
+			    nredundant == 1 ? "" : "s");
+		} else {
+			printf(", %d pack file%s", nredundant,
+			    nredundant == 1 ? "" : "s");
+		}
+	}
+	if (print_loose || print_commits || print_purged || print_redundant) {
 		a->printed_something = 1;
 		fflush(stdout);
 	}
@@ -1217,18 +1268,24 @@ cmd_cleanup(int argc, char *argv[])
 	const struct got_error *error = NULL;
 	char *repo_path = NULL;
 	struct got_repository *repo = NULL;
-	int ch, dry_run = 0, npacked = 0, verbosity = 0;
+	int ch, dry_run = 0, verbosity = 0;
+	int ncommits = 0, nloose = 0, npacked = 0;
 	int remove_lonely_packidx = 0, ignore_mtime = 0;
+	struct got_pack_progress_arg ppa;
 	struct got_cleanup_progress_arg cpa;
 	struct got_lonely_packidx_progress_arg lpa;
-	off_t size_before, size_after;
-	char scaled_before[FMT_SCALED_STRSIZE];
-	char scaled_after[FMT_SCALED_STRSIZE];
-	char scaled_diff[FMT_SCALED_STRSIZE];
+	off_t loose_before, loose_after;
+	off_t pack_before, pack_after;
+	off_t total_size;
+	char loose_before_scaled[FMT_SCALED_STRSIZE];
+	char loose_after_scaled[FMT_SCALED_STRSIZE];
+	char pack_before_scaled[FMT_SCALED_STRSIZE];
+	char pack_after_scaled[FMT_SCALED_STRSIZE];
+	char total_size_scaled[FMT_SCALED_STRSIZE];
 	int *pack_fds = NULL;
 
 #ifndef PROFILE
-	if (pledge("stdio rpath wpath cpath flock proc exec sendfd unveil",
+	if (pledge("stdio rpath wpath cpath fattr flock proc exec sendfd unveil",
 	    NULL) == -1)
 		err(1, "pledge");
 #endif
@@ -1296,39 +1353,62 @@ cmd_cleanup(int argc, char *argv[])
 	}
 
 	memset(&cpa, 0, sizeof(cpa));
-	cpa.last_ncommits = -1;
+	cpa.last_nloose = -1;
 	cpa.last_npurged = -1;
+	cpa.last_nredundant = -1;
 	cpa.dry_run = dry_run;
 	cpa.verbosity = verbosity;
-	error = got_repo_purge_unreferenced_loose_objects(repo,
-	    &size_before, &size_after, &npacked, dry_run, ignore_mtime,
-	    cleanup_progress, &cpa, check_cancelled, NULL);
-	if (cpa.printed_something)
+
+	memset(&ppa, 0, sizeof(ppa));
+	ppa.out = stdout;
+	ppa.verbosity = verbosity;
+
+	error = got_repo_cleanup(repo, &loose_before, &loose_after,
+	    &pack_before, &pack_after, &ncommits, &nloose, &npacked,
+	    dry_run, ignore_mtime, cleanup_progress, &cpa,
+	    pack_progress, &ppa, pack_index_progress, &ppa,
+	    check_cancelled, NULL);
+	if (ppa.printed_something || cpa.printed_something)
 		printf("\n");
 	if (error)
 		goto done;
+
+	total_size = (loose_before - loose_after) + (pack_before - pack_after);
+
 	if (cpa.printed_something) {
-		if (fmt_scaled(size_before, scaled_before) == -1) {
+		if (fmt_scaled(loose_before, loose_before_scaled) == -1) {
 			error = got_error_from_errno("fmt_scaled");
 			goto done;
 		}
-		if (fmt_scaled(size_after, scaled_after) == -1) {
+		if (fmt_scaled(loose_after, loose_after_scaled) == -1) {
 			error = got_error_from_errno("fmt_scaled");
 			goto done;
 		}
-		if (fmt_scaled(size_before - size_after, scaled_diff) == -1) {
+		if (fmt_scaled(pack_before, pack_before_scaled) == -1) {
 			error = got_error_from_errno("fmt_scaled");
 			goto done;
 		}
-		printf("loose total size before: %s\n", scaled_before);
-		printf("loose total size after: %s\n", scaled_after);
+		if (fmt_scaled(pack_after, pack_after_scaled) == -1) {
+			error = got_error_from_errno("fmt_scaled");
+			goto done;
+		}
+		if (fmt_scaled(total_size, total_size_scaled) == -1) {
+			error = got_error_from_errno("fmt_scaled");
+			goto done;
+		}
+		printf("loose total size before: %s\n", loose_before_scaled);
+		printf("loose total size after: %s\n", loose_after_scaled);
+		printf("pack files total size before: %s\n",
+		    pack_before_scaled);
+		printf("pack files total size after: %s\n", pack_after_scaled);
 		if (dry_run) {
 			printf("disk space which would be freed: %s\n",
-			    scaled_diff);
+			    total_size_scaled);
 		} else
-			printf("disk space freed: %s\n", scaled_diff);
+			printf("disk space freed: %s\n", total_size_scaled);
 		printf("loose objects also found in pack files: %d\n", npacked);
 	}
+
 done:
 	if (repo)
 		got_repo_close(repo);
@@ -1341,3 +1421,428 @@ done:
 	free(repo_path);
 	return error;
 }
+
+__dead static void
+usage_dump(void)
+{
+	fprintf(stderr, "usage: %s dump [-q] [-r repository-path] "
+	    "[-x reference] [reference]...\n", getprogname());
+	exit(1);
+}
+
+static const struct got_error *
+cmd_dump(int argc, char *argv[])
+{
+	const struct got_error *error = NULL;
+	struct got_pack_progress_arg ppa;
+	struct got_repository *repo = NULL;
+	struct got_pathlist_head exclude_args;
+	struct got_pathlist_entry *pe;
+	struct got_reflist_head exclude_refs;
+	struct got_reflist_head include_refs;
+	struct got_reflist_entry *re, *new;
+	const char *refname;
+	char *repo_path = NULL;
+	int *pack_fds = NULL;
+	int verbosity = 0;
+	int i, ch;
+
+	RB_INIT(&exclude_args);
+	TAILQ_INIT(&exclude_refs);
+	TAILQ_INIT(&include_refs);
+
+#ifndef PROFILE
+	if (pledge("stdio rpath wpath cpath flock proc exec sendfd unveil",
+	    NULL) == -1)
+		err(1, "pledge");
+#endif
+
+	while ((ch = getopt(argc, argv, "qr:x:")) != -1) {
+		switch (ch) {
+		case 'q':
+			verbosity = -1;
+			break;
+		case 'r':
+			repo_path = realpath(optarg, NULL);
+			if (repo_path == NULL)
+				return got_error_from_errno2("realpath",
+				    optarg);
+			got_path_strip_trailing_slashes(repo_path);
+			break;
+		case 'x':
+			error = got_pathlist_insert(NULL, &exclude_args,
+			    optarg, NULL);
+			if (error)
+				return error;
+			break;
+		default:
+			usage_dump();
+			/* NOTREACHED */
+		}
+	}
+	argc -= optind;
+	argv += optind;
+
+	if (repo_path == NULL) {
+		error = get_repo_path(&repo_path);
+		if (error)
+			goto done;
+	}
+	error = got_repo_pack_fds_open(&pack_fds);
+	if (error != NULL)
+		goto done;
+	error = got_repo_open(&repo, repo_path, NULL, pack_fds);
+	if (error)
+		goto done;
+
+	error = apply_unveil(got_repo_get_path_git_dir(repo), 1);
+	if (error)
+		goto done;
+
+	RB_FOREACH(pe, got_pathlist_head, &exclude_args) {
+		refname = pe->path;
+		error = add_ref(&new, &exclude_refs, refname, repo);
+		if (error)
+			goto done;
+	}
+
+	if (argc == 0) {
+		error = got_ref_list(&include_refs, repo, "",
+		    got_ref_cmp_by_name, NULL);
+		if (error)
+			goto done;
+	} else {
+		for (i = 0; i < argc; i++) {
+			got_path_strip_trailing_slashes(argv[i]);
+			refname = argv[i];
+			error = add_ref(&new, &include_refs, refname, repo);
+			if (error)
+				goto done;
+		}
+	}
+
+	/* Ignore references in the refs/got/ namespace. */
+	TAILQ_FOREACH_SAFE(re, &include_refs, entry, new) {
+		refname = got_ref_get_name(re->ref);
+		if (strncmp("refs/got/", refname, 9) != 0)
+			continue;
+		TAILQ_REMOVE(&include_refs, re, entry);
+		got_ref_close(re->ref);
+		free(re);
+	}
+
+	memset(&ppa, 0, sizeof(ppa));
+	ppa.out = stderr;
+	ppa.verbosity = verbosity;
+
+	error = got_repo_dump(stdout, &include_refs, &exclude_refs,
+	    repo, pack_progress, &ppa, check_cancelled, NULL);
+	if (ppa.printed_something)
+		fprintf(stderr, "\n");
+ done:
+	if (repo)
+		got_repo_close(repo);
+
+	if (pack_fds) {
+		const struct got_error *pack_err;
+
+		pack_err = got_repo_pack_fds_close(pack_fds);
+		if (error == NULL)
+			error = pack_err;
+	}
+
+	got_pathlist_free(&exclude_args, GOT_PATHLIST_FREE_NONE);
+	got_ref_list_free(&exclude_refs);
+	got_ref_list_free(&include_refs);
+	free(repo_path);
+
+	return error;
+}
+
+__dead static void
+usage_load(void)
+{
+	fprintf(stderr, "usage: %s load [-nq] [-l bundle-file] "
+	    "[-r repository-path] [reference ...]\n",
+	    getprogname());
+	exit(1);
+}
+
+static const struct got_error *
+load_progress(void *arg, off_t packfile_size, int nobj_total,
+    int nobj_indexed, int nobj_loose, int nobj_resolved)
+{
+	return pack_index_progress(arg, packfile_size, nobj_total,
+	    nobj_indexed, nobj_loose, nobj_resolved, 0);
+}
+
+static int
+is_wanted_ref(struct got_pathlist_head *wanted, const char *ref)
+{
+	struct got_pathlist_entry *pe;
+
+	if (RB_EMPTY(wanted))
+		return 1;
+
+	RB_FOREACH(pe, got_pathlist_head, wanted) {
+		if (strcmp(pe->path, ref) == 0)
+			return 1;
+	}
+
+	return 0;
+}
+
+static const struct got_error *
+create_ref(const char *refname, struct got_object_id *id,
+    int verbosity, struct got_repository *repo)
+{
+	const struct got_error *err = NULL;
+	struct got_reference *ref;
+	char *id_str;
+
+	err = got_object_id_str(&id_str, id);
+	if (err)
+		return err;
+
+	err = got_ref_alloc(&ref, refname, id);
+	if (err)
+		goto done;
+
+	err = got_ref_write(ref, repo);
+	got_ref_close(ref);
+
+	if (err == NULL && verbosity >= 0)
+		printf("Created reference %s: %s\n", refname, id_str);
+done:
+	free(id_str);
+	return err;
+}
+
+static const struct got_error *
+update_ref(struct got_reference *ref, struct got_object_id *new_id,
+    int replace_tags, int verbosity, struct got_repository *repo)
+{
+	const struct got_error *err = NULL;
+	char *new_id_str = NULL;
+	struct got_object_id *old_id = NULL;
+
+	err = got_object_id_str(&new_id_str, new_id);
+	if (err)
+		goto done;
+
+	if (!replace_tags &&
+	    strncmp(got_ref_get_name(ref), "refs/tags/", 10) == 0) {
+		err = got_ref_resolve(&old_id, repo, ref);
+		if (err)
+			goto done;
+		if (got_object_id_cmp(old_id, new_id) == 0)
+			goto done;
+		if (verbosity >= 0) {
+			printf("Rejecting update of existing tag %s: %s\n",
+			    got_ref_get_name(ref), new_id_str);
+		}
+		goto done;
+	}
+
+	if (got_ref_is_symbolic(ref)) {
+		if (verbosity >= 0) {
+			printf("Replacing reference %s: %s\n",
+			    got_ref_get_name(ref),
+			    got_ref_get_symref_target(ref));
+		}
+		err = got_ref_change_symref_to_ref(ref, new_id);
+		if (err)
+			goto done;
+		err = got_ref_write(ref, repo);
+		if (err)
+			goto done;
+	} else {
+		err = got_ref_resolve(&old_id, repo, ref);
+		if (err)
+			goto done;
+		if (got_object_id_cmp(old_id, new_id) == 0)
+			goto done;
+
+		err = got_ref_change_ref(ref, new_id);
+		if (err)
+			goto done;
+		err = got_ref_write(ref, repo);
+		if (err)
+			goto done;
+	}
+
+	if (verbosity >= 0)
+		printf("Updated %s: %s\n", got_ref_get_name(ref),
+		    new_id_str);
+done:
+	free(old_id);
+	free(new_id_str);
+	return err;
+}
+
+static const struct got_error *
+cmd_load(int argc, char *argv[])
+{
+	const struct got_error *error = NULL;
+	struct got_repository *repo = NULL;
+	struct got_pathlist_head include_args;
+	struct got_pathlist_head available_refs;
+	struct got_pathlist_entry *pe;
+	struct got_pack_progress_arg ppa;
+	FILE *in = stdin;
+	int *pack_fds = NULL;
+	char *repo_path = NULL;
+	int list_refs_only = 0;
+	int noop = 0;
+	int verbosity = 0;
+	int ch, i;
+
+	RB_INIT(&include_args);
+	RB_INIT(&available_refs);
+
+#ifndef PROFILE
+	if (pledge("stdio rpath wpath cpath fattr flock proc exec "
+	    "sendfd unveil", NULL) == -1)
+		err(1, "pledge");
+#endif
+
+	while ((ch = getopt(argc, argv, "l:nqr:")) != -1) {
+		switch (ch) {
+		case 'l':
+			list_refs_only = 1;
+			in = fopen(optarg, "re");
+			if (in == NULL)
+				return got_error_from_errno2("open", optarg);
+			break;
+		case 'n':
+			noop = 1;
+			break;
+		case 'q':
+			verbosity = -1;
+			break;
+		case 'r':
+			repo_path = realpath(optarg, NULL);
+			if (repo_path == NULL)
+				return got_error_from_errno2("realpath",
+				    optarg);
+			got_path_strip_trailing_slashes(repo_path);
+			break;
+		default:
+			usage_load();
+			/* NOTREACHED */
+		}
+	}
+	argc -= optind;
+	argv += optind;
+
+	if (list_refs_only && argc > 1)
+		errx(1, "-l and references on the command line are exclusive");
+	if (list_refs_only && noop)
+		errx(1, "-n and -l are mutually exclusive");
+
+	for (i = 0; i < argc; i++) {
+		char *refname = argv[i];
+		got_path_strip_trailing_slashes(refname);
+		if (!got_ref_name_is_valid(refname))
+			errx(1, "invalid reference name %s", refname);
+		error = got_pathlist_insert(NULL, &include_args, refname, NULL);
+		if (error)
+			goto done;
+	}
+
+	if (repo_path == NULL) {
+		error = get_repo_path(&repo_path);
+		if (error)
+			goto done;
+	}
+	error = got_repo_pack_fds_open(&pack_fds);
+	if (error != NULL)
+		goto done;
+	error = got_repo_open(&repo, repo_path, NULL, pack_fds);
+	if (error)
+		goto done;
+
+	error = apply_unveil(got_repo_get_path_git_dir(repo), 0);
+	if (error)
+		goto done;
+
+	memset(&ppa, 0, sizeof(ppa));
+	ppa.out = stdout;
+	ppa.verbosity = verbosity;
+
+	error = got_repo_load(in, &available_refs, repo, list_refs_only, noop,
+	    load_progress, &ppa, check_cancelled, NULL);
+	if (verbosity >= 0 && !list_refs_only)
+		printf("\n");
+	if (error)
+		goto done;
+
+	if (list_refs_only) {
+		RB_FOREACH(pe, got_pathlist_head, &available_refs) {
+			const char *refname = pe->path;
+			struct got_object_id *id = pe->data;
+			char *idstr;
+
+			error = got_object_id_str(&idstr, id);
+			if (error)
+				goto done;
+
+			printf("%s: %s\n", refname, idstr);
+			free(idstr);
+		}
+		goto done;
+	}
+
+	if (noop)
+		goto done;
+
+	/* Update references */
+	RB_FOREACH(pe, got_pathlist_head, &available_refs) {
+		const struct got_error *unlock_err;
+		struct got_reference *ref;
+		const char *refname = pe->path;
+		struct got_object_id *id = pe->data;
+
+		if (!is_wanted_ref(&include_args, pe->path))
+			continue;
+
+		error = got_ref_open(&ref, repo, refname, 1);
+		if (error) {
+			if (error->code != GOT_ERR_NOT_REF)
+				goto done;
+			error = create_ref(refname, id, verbosity, repo);
+			if (error)
+				goto done;
+		} else {
+			/* XXX: check advances only and add -f to force? */
+			error = update_ref(ref, id, 1, verbosity, repo);
+			unlock_err = got_ref_unlock(ref);
+			if (unlock_err && error == NULL)
+				error = unlock_err;
+			got_ref_close(ref);
+			if (error)
+				goto done;
+		}
+	}
+
+ done:
+	if (in != stdin && fclose(in) == EOF && error == NULL)
+		error = got_error_from_errno("fclose");
+
+	if (repo)
+		got_repo_close(repo);
+
+	if (pack_fds) {
+		const struct got_error *pack_err;
+
+		pack_err = got_repo_pack_fds_close(pack_fds);
+		if (error == NULL)
+			error = pack_err;
+	}
+
+	got_pathlist_free(&include_args, GOT_PATHLIST_FREE_NONE);
+	got_pathlist_free(&available_refs, GOT_PATHLIST_FREE_ALL);
+	free(repo_path);
+
+	return error;
+}
