@@ -22,9 +22,10 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sha1.h>
+#include <sha2.h>
 #include <zlib.h>
-
-#include "got_compat.h"
+#include <uuid.h>
 
 #include "got_error.h"
 #include "got_object.h"
@@ -39,17 +40,8 @@
 #define nitems(_a) (sizeof(_a) / sizeof((_a)[0]))
 #endif
 
-#if defined(__GLIBC__)
-	/*
-	 * The autoconf test for strerror_r is broken in current versions
-	 * of autoconf: https://savannah.gnu.org/support/?110367
-	 */
-char *__xpg_strerror_r(int, char *, size_t);
-#define strerror_r __xpg_strerror_r
-#endif
-
 static const struct got_error got_errors[] = {
-	{ GOT_ERR_OK,		"no error occured?!?" },
+	{ GOT_ERR_OK,		"no error occurred?!?" },
 	{ GOT_ERR_ERRNO,	"see errno" },
 	{ GOT_ERR_NOT_GIT_REPO, "no git repository found" },
 	{ GOT_ERR_BAD_FILETYPE,	"bad file type" },
@@ -91,7 +83,7 @@ static const struct got_error got_errors[] = {
 	{ GOT_ERR_PACK_OFFSET,	"bad offset in pack file" },
 	{ GOT_ERR_OBJ_EXISTS,	"object already exists" },
 	{ GOT_ERR_BAD_OBJ_ID,	"bad object id" },
-	{ GOT_ERR_ITER_BUSY,	"iteration already in progress" },
+	{ GOT_ERR_OBJECT_FORMAT, "object format not supported" },
 	{ GOT_ERR_ITER_COMPLETED,"iteration completed" },
 	{ GOT_ERR_RANGE,	"value out of range" },
 	{ GOT_ERR_EXPECTED,	"expected an error but have no error" },
@@ -140,6 +132,7 @@ static const struct got_error got_errors[] = {
 	    "changes must be committed or reverted first" },
 	{ GOT_ERR_NOT_REBASING,	"rebase operation not in progress" },
 	{ GOT_ERR_REBASE_COMMITID,"rebase commit ID mismatch" },
+	{ GOT_ERR_WRONG_BRANCH, "update -b required" },
 	{ GOT_ERR_REBASING,	"a rebase operation is in progress in this "
 	    "work tree and must be continued or aborted first" },
 	{ GOT_ERR_REBASE_PATH,	"cannot rebase branch which contains "
@@ -200,7 +193,7 @@ static const struct got_error got_errors[] = {
 	{ GOT_ERR_SEND_BAD_REF, "reference cannot be sent" },
 	{ GOT_ERR_SEND_FAILED, "could not send pack file" },
 	{ GOT_ERR_SEND_EMPTY, "no references to send" },
-	{ GOT_ERR_SEND_ANCESTRY, "fetch and rebase required" },
+	{ GOT_ERR_SEND_ANCESTRY, "branch on server has a different ancestry; either fetch changes from server and then rebase or merge local branch before sending, or ignore ancestry with send -f (can lead to data loss on server)" },
 	{ GOT_ERR_CAPA_DELETE_REFS, "server cannot delete references" },
 	{ GOT_ERR_SEND_DELETE_REF, "reference cannot be deleted" },
 	{ GOT_ERR_SEND_TAG_EXISTS, "tag already exists on server" },
@@ -244,7 +237,19 @@ static const struct got_error got_errors[] = {
 	    "the work tree is no longer up-to-date; merge must be aborted "
 	    "and retried" },
 	{ GOT_ERR_BUNDLE_FORMAT, "unknown git bundle version" },
-	{ GOT_ERR_BAD_KEYWORD, "invalid commit keyword" }
+	{ GOT_ERR_BAD_KEYWORD, "invalid commit keyword" },
+	{ GOT_ERR_UNKNOWN_CAPA, "unknown capability" },
+	{ GOT_ERR_REF_DUP_ENTRY, "duplicate reference entry" },
+	{ GOT_ERR_DIFF_NOCHANGES, "no changes match the requested diff" },
+	{ GOT_ERR_USER, "no such user" },
+	{ GOT_ERR_USER_EXISTS, "user already exists" },
+	{ GOT_ERR_GROUP, "no such group" },
+	{ GOT_ERR_GROUP_EXISTS, "group already exists" },
+	{ GOT_ERR_AUTHORIZED_KEY, "no authorized key found" },
+	{ GOT_ERR_CONNECTION_LIMIT, "connection limit exceeded" },
+	{ GOT_ERR_ON_SERVER_SIDE, "see server-side logs for error details" },
+	{ GOT_ERR_LOGIN_FAILED, "login failed" },
+	{ GOT_ERR_UNKNOWN_COMMAND, "command not found" },
 };
 
 static struct got_custom_error {
