@@ -14,19 +14,21 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-#include "got_compat.h"
-
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/queue.h>
+#include <sys/tree.h>
 #include <sys/uio.h>
 #include <sys/mman.h>
 
+#include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sha1.h>
+#include <sha2.h>
 #include <unistd.h>
 #include <zlib.h>
 #include <libgen.h>
@@ -115,7 +117,7 @@ got_object_get_path(char **path, struct got_object_id *id,
 		goto done;
 
 	if (asprintf(path, "%s/%.2x/%s", path_objects,
-	    id->sha1[0], hex + 2) == -1)
+	    id->hash[0], hex + 2) == -1)
 		err = got_error_from_errno("asprintf");
 
 done:
@@ -136,7 +138,10 @@ got_object_open_loose_fd(int *fd, struct got_object_id *id,
 		return err;
 	*fd = open(path, O_RDONLY | O_NOFOLLOW | O_CLOEXEC);
 	if (*fd == -1) {
-		err = got_error_from_errno2("open", path);
+		if (errno == ENOENT)
+			err = got_error_no_obj(id);
+		else
+			err = got_error_from_errno2("open", path);
 		goto done;
 	}
 done:
@@ -150,7 +155,7 @@ got_object_open_by_id_str(struct got_object **obj, struct got_repository *repo,
 {
 	struct got_object_id id;
 
-	if (!got_parse_object_id(&id, id_str, GOT_HASH_SHA1))
+	if (!got_parse_object_id(&id, id_str, repo->algo))
 		return got_error_path(id_str, GOT_ERR_BAD_OBJ_ID_STR);
 
 	return got_object_open(obj, repo, &id);
@@ -175,41 +180,6 @@ got_object_resolve_id_str(struct got_object_id **id,
 	return NULL;
 }
 
-const struct got_error *
-got_object_qid_alloc(struct got_object_qid **qid, struct got_object_id *id)
-{
-	*qid = calloc(1, sizeof(**qid));
-	if (*qid == NULL)
-		return got_error_from_errno("calloc");
-
-	memcpy(&(*qid)->id, id, sizeof((*qid)->id));
-	return NULL;
-}
-
-const struct got_error *
-got_object_id_queue_copy(const struct got_object_id_queue *src,
-    struct got_object_id_queue *dest)
-{
-	const struct got_error *err;
-	struct got_object_qid *qid;
-
-	STAILQ_FOREACH(qid, src, entry) {
-		struct got_object_qid *new;
-		/*
-		 * Deep-copy the object ID only. Let the caller deal
-		 * with setting up the new->data pointer if needed.
-		 */
-		err = got_object_qid_alloc(&new, &qid->id);
-		if (err) {
-			got_object_id_queue_free(dest);
-			return err;
-		}
-		STAILQ_INSERT_TAIL(dest, new, entry);
-	}
-
-	return NULL;
-}
-
 int
 got_object_tree_get_nentries(struct got_tree_object *tree)
 {
@@ -405,7 +375,7 @@ const struct got_error *
 got_object_blob_is_binary(int *binary, struct got_blob_object *blob)
 {
 	const struct got_error *err;
-	size_t hdrlen, len;
+	size_t hdrlen, len, i;
 
 	*binary = 0;
 	hdrlen = got_object_blob_get_hdrlen(blob);
@@ -417,7 +387,13 @@ got_object_blob_is_binary(int *binary, struct got_blob_object *blob)
 	if (err)
 		return err;
 
-	*binary = memchr(blob->read_buf, '\0', len) != NULL;
+	for (i = 0; i < len; ++i) {
+		if (iscntrl((unsigned char)blob->read_buf[i]) &&
+		    !isspace((unsigned char)blob->read_buf[i])) {
+			*binary = 1;
+			break;
+		}
+	}
 
 	if (fseeko(blob->f, hdrlen, SEEK_SET) == -1)
 		return got_error_from_errno("fseeko");
