@@ -17,6 +17,7 @@
 
 #include <sys/types.h>
 #include <sys/queue.h>
+#include <sys/tree.h>
 #include <sys/uio.h>
 #include <sys/wait.h>
 
@@ -30,11 +31,11 @@
 #include <stdint.h>
 #include <poll.h>
 #include <imsg.h>
+#include <sha1.h>
+#include <sha2.h>
 #include <unistd.h>
 #include <zlib.h>
 
-#include "got_compat.h"
-
 #include "got_object.h"
 #include "got_error.h"
 #include "got_path.h"
@@ -45,6 +46,7 @@
 #include "got_lib_inflate.h"
 #include "got_lib_object.h"
 #include "got_lib_object_parse.h"
+#include "got_lib_object_qid.h"
 #include "got_lib_privsep.h"
 #include "got_lib_pack.h"
 #include "got_lib_poll.h"
@@ -72,12 +74,9 @@ read_imsg(struct imsgbuf *ibuf)
 		return err;
 	}
 
-	n = imsg_read(ibuf);
-	if (n == -1) {
-		if (errno == EAGAIN) /* Could be a file-descriptor leak. */
-			return got_error(GOT_ERR_PRIVSEP_NO_FD);
+	n = imsgbuf_read(ibuf);
+	if (n == -1)
 		return got_error(GOT_ERR_PRIVSEP_READ);
-	}
 	if (n == 0)
 		return got_error(GOT_ERR_PRIVSEP_PIPE);
 
@@ -140,12 +139,16 @@ got_privsep_recv_imsg(struct imsg *imsg, struct imsgbuf *ibuf,
 			return got_error_from_errno("imsg_get");
 	}
 
-	if (imsg->hdr.len < IMSG_HEADER_SIZE + min_datalen)
-		return got_error(GOT_ERR_PRIVSEP_LEN);
-
 	if (imsg->hdr.type == GOT_IMSG_ERROR) {
 		size_t datalen = imsg->hdr.len - IMSG_HEADER_SIZE;
-		return recv_imsg_error(imsg, datalen);
+		err = recv_imsg_error(imsg, datalen);
+		imsg_free(imsg);
+		return err;
+	}
+
+	if (imsg->hdr.len < IMSG_HEADER_SIZE + min_datalen) {
+		imsg_free(imsg);
+		return got_error(GOT_ERR_PRIVSEP_LEN);
 	}
 
 	return NULL;
@@ -173,16 +176,16 @@ got_privsep_send_error(struct imsgbuf *ibuf, const struct got_error *err)
 
 	poll_err = got_poll_fd(ibuf->fd, POLLOUT, INFTIM);
 	if (poll_err) {
-		fprintf(stderr, "%s: error %d \"%s\": poll: %s\n",
-		    getprogname(), err->code, err->msg, poll_err->msg);
+		if (poll_err->code != GOT_ERR_EOF)
+			fprintf(stderr, "%s: error %d \"%s\": poll: %s\n",
+			    getprogname(), err->code, err->msg, poll_err->msg);
 		return;
 	}
 
-	ret = imsg_flush(ibuf);
+	ret = imsgbuf_flush(ibuf);
 	if (ret == -1) {
-		fprintf(stderr, "%s: error %d \"%s\": imsg_flush: %s\n",
+		fprintf(stderr, "%s: error %d \"%s\": imsgbuf_flush: %s\n",
 		    getprogname(), err->code, err->msg, strerror(errno));
-		imsg_clear(ibuf);
 		return;
 	}
 }
@@ -196,10 +199,8 @@ flush_imsg(struct imsgbuf *ibuf)
 	if (err)
 		return err;
 
-	if (imsg_flush(ibuf) == -1) {
-		imsg_clear(ibuf);
-		return got_error_from_errno("imsg_flush");
-	}
+	if (imsgbuf_flush(ibuf) == -1)
+		return got_error_from_errno("imsgbuf_flush");
 
 	return NULL;
 }
@@ -216,12 +217,17 @@ got_privsep_send_stop(int fd)
 	const struct got_error *err = NULL;
 	struct imsgbuf ibuf;
 
-	imsg_init(&ibuf, fd);
+	if (imsgbuf_init(&ibuf, fd) == -1)
+		return got_error_from_errno("imsgbuf_init");
 
-	if (imsg_compose(&ibuf, GOT_IMSG_STOP, 0, 0, -1, NULL, 0) == -1)
-		return got_error_from_errno("imsg_compose STOP");
+	if (imsg_compose(&ibuf, GOT_IMSG_STOP, 0, 0, -1, NULL, 0) == -1) {
+		err = got_error_from_errno("imsg_compose STOP");
+		goto done;
+	}
 
 	err = flush_imsg(&ibuf);
+ done:
+	imsgbuf_clear(&ibuf);
 	return err;
 }
 
@@ -240,9 +246,14 @@ const struct got_error *
 got_privsep_send_raw_obj_req(struct imsgbuf *ibuf, int fd,
     struct got_object_id *id)
 {
+	const struct got_error *err;
+
 	if (imsg_compose(ibuf, GOT_IMSG_RAW_OBJECT_REQUEST, 0, 0, fd,
-	    id, sizeof(*id)) == -1)
-		return got_error_from_errno("imsg_compose RAW_OBJECT_REQUEST");
+	    id, sizeof(*id)) == -1) {
+		err = got_error_from_errno("imsg_compose RAW_OBJECT_REQUEST");
+		close(fd);
+		return err;
+	}
 
 	return flush_imsg(ibuf);
 }
@@ -266,7 +277,6 @@ const struct got_error *
 got_privsep_send_raw_obj(struct imsgbuf *ibuf, off_t size, size_t hdrlen,
     uint8_t *data)
 {
-	const struct got_error *err = NULL;
 	struct got_imsg_raw_obj iobj;
 	size_t len = sizeof(iobj);
 	struct ibuf *wbuf;
@@ -279,10 +289,8 @@ got_privsep_send_raw_obj(struct imsgbuf *ibuf, off_t size, size_t hdrlen,
 		len += (size_t)size + hdrlen;
 
 	wbuf = imsg_create(ibuf, GOT_IMSG_RAW_OBJECT, 0, 0, len);
-	if (wbuf == NULL) {
-		err = got_error_from_errno("imsg_create RAW_OBJECT");
-		return err;
-	}
+	if (wbuf == NULL)
+		return got_error_from_errno("imsg_create RAW_OBJECT");
 
 	if (imsg_add(wbuf, &iobj, sizeof(iobj)) == -1)
 		return got_error_from_errno("imsg_add RAW_OBJECT");
@@ -292,9 +300,7 @@ got_privsep_send_raw_obj(struct imsgbuf *ibuf, off_t size, size_t hdrlen,
 			return got_error_from_errno("imsg_add RAW_OBJECT");
 	}
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
-
 	return flush_imsg(ibuf);
 }
 
@@ -387,6 +393,7 @@ const struct got_error *
 got_privsep_send_tree_req(struct imsgbuf *ibuf, int fd,
     struct got_object_id *id, int pack_idx)
 {
+	const struct got_error *err;
 	struct ibuf *wbuf;
 	size_t len;
 
@@ -396,18 +403,30 @@ got_privsep_send_tree_req(struct imsgbuf *ibuf, int fd,
 		len = sizeof(*id);
 
 	wbuf = imsg_create(ibuf, GOT_IMSG_TREE_REQUEST, 0, 0, len);
-	if (wbuf == NULL)
-		return got_error_from_errno("imsg_create TREE_REQUEST");
-
-	if (imsg_add(wbuf, id, sizeof(*id)) == -1)
-		return got_error_from_errno("imsg_add TREE_REQUEST");
-
-	if (pack_idx != -1) { /* tree is packed */
-		if (imsg_add(wbuf, &pack_idx, sizeof(pack_idx)) == -1)
-			return got_error_from_errno("imsg_add TREE_REQUEST");
+	if (wbuf == NULL) {
+		err = got_error_from_errno("imsg_create TREE_REQUEST");
+		if (fd != -1)
+			close(fd);
+		return err;
 	}
 
-	wbuf->fd = fd;
+	if (imsg_add(wbuf, id, sizeof(*id)) == -1) {
+		err = got_error_from_errno("imsg_add TREE_REQUEST");
+		if (fd != -1)
+			close(fd);
+		return err;
+	}
+
+	if (pack_idx != -1) { /* tree is packed */
+		if (imsg_add(wbuf, &pack_idx, sizeof(pack_idx)) == -1) {
+			err = got_error_from_errno("imsg_add TREE_REQUEST");
+			if (fd != -1)
+				close(fd);
+			return err;
+		}
+	}
+
+	ibuf_fd_set(wbuf, fd);
 	imsg_close(ibuf, wbuf);
 
 	return flush_imsg(ibuf);
@@ -417,6 +436,7 @@ const struct got_error *
 got_privsep_send_tag_req(struct imsgbuf *ibuf, int fd,
     struct got_object_id *id, int pack_idx)
 {
+	const struct got_error *err;
 	struct got_imsg_packed_object iobj;
 	void *data;
 	size_t len;
@@ -433,8 +453,12 @@ got_privsep_send_tag_req(struct imsgbuf *ibuf, int fd,
 	}
 
 	if (imsg_compose(ibuf, GOT_IMSG_TAG_REQUEST, 0, 0, fd, data, len)
-	    == -1)
-		return got_error_from_errno("imsg_compose TAG_REQUEST");
+	    == -1) {
+		err = got_error_from_errno("imsg_compose TAG_REQUEST");
+		if (fd != -1)
+			close(fd);
+		return err;
+	}
 
 	return flush_imsg(ibuf);
 }
@@ -557,8 +581,11 @@ got_privsep_send_fetch_req(struct imsgbuf *ibuf, int fd,
 	}
 
 	wbuf = imsg_create(ibuf, GOT_IMSG_FETCH_REQUEST, 0, 0, len);
-	if (wbuf == NULL)
-		return got_error_from_errno("imsg_create FETCH_HAVE_REF");
+	if (wbuf == NULL) {
+		err = got_error_from_errno("imsg_create FETCH_HAVE_REF");
+		close(fd);
+		return err;
+	}
 
 	memset(&fetchreq, 0, sizeof(fetchreq));
 	fetchreq.no_head = no_head;
@@ -569,33 +596,38 @@ got_privsep_send_fetch_req(struct imsgbuf *ibuf, int fd,
 		fetchreq.worktree_branch_len = worktree_branch_len;
 	if (remote_head != NULL)
 		fetchreq.remote_head_len = remote_head_len;
-	TAILQ_FOREACH(pe, have_refs, entry)
+	RB_FOREACH(pe, got_pathlist_head, have_refs)
 		fetchreq.n_have_refs++;
-	TAILQ_FOREACH(pe, wanted_branches, entry)
+	RB_FOREACH(pe, got_pathlist_head, wanted_branches)
 		fetchreq.n_wanted_branches++;
-	TAILQ_FOREACH(pe, wanted_refs, entry)
+	RB_FOREACH(pe, got_pathlist_head, wanted_refs)
 		fetchreq.n_wanted_refs++;
 	if (imsg_add(wbuf, &fetchreq, sizeof(fetchreq)) == -1)
 		return got_error_from_errno("imsg_add FETCH_REQUEST");
 	if (worktree_branch) {
-		if (imsg_add(wbuf, worktree_branch, worktree_branch_len) == -1)
-			return got_error_from_errno("imsg_add FETCH_REQUEST");
+		if (imsg_add(wbuf, worktree_branch, worktree_branch_len)
+		    == -1) {
+			err = got_error_from_errno("imsg_add FETCH_REQUEST");
+			close(fd);
+			return err;
+		}
 	}
 	if (remote_head) {
-		if (imsg_add(wbuf, remote_head, remote_head_len) == -1)
-			return got_error_from_errno("imsg_add FETCH_REQUEST");
+		if (imsg_add(wbuf, remote_head, remote_head_len) == -1) {
+			err = got_error_from_errno("imsg_add FETCH_REQUEST");
+			close(fd);
+			return err;
+		}
 	}
-	wbuf->fd = fd;
+	ibuf_fd_set(wbuf, fd);
+	fd = -1;
 	imsg_close(ibuf, wbuf);
 
 	err = flush_imsg(ibuf);
-	if (err) {
-		close(fd);
+	if (err)
 		return err;
-	}
-	fd = -1;
 
-	TAILQ_FOREACH(pe, have_refs, entry) {
+	RB_FOREACH(pe, got_pathlist_head, have_refs) {
 		const char *name = pe->path;
 		size_t name_len = pe->path_len;
 		struct got_object_id *id = pe->data;
@@ -613,14 +645,13 @@ got_privsep_send_fetch_req(struct imsgbuf *ibuf, int fd,
 		if (imsg_add(wbuf, name, name_len) == -1)
 			return got_error_from_errno("imsg_add FETCH_HAVE_REF");
 
-		wbuf->fd = -1;
 		imsg_close(ibuf, wbuf);
 		err = flush_imsg(ibuf);
 		if (err)
 			return err;
 	}
 
-	TAILQ_FOREACH(pe, wanted_branches, entry) {
+	RB_FOREACH(pe, got_pathlist_head, wanted_branches) {
 		const char *name = pe->path;
 		size_t name_len = pe->path_len;
 
@@ -639,14 +670,13 @@ got_privsep_send_fetch_req(struct imsgbuf *ibuf, int fd,
 			return got_error_from_errno(
 			    "imsg_add FETCH_WANTED_BRANCH");
 
-		wbuf->fd = -1;
 		imsg_close(ibuf, wbuf);
 		err = flush_imsg(ibuf);
 		if (err)
 			return err;
 	}
 
-	TAILQ_FOREACH(pe, wanted_refs, entry) {
+	RB_FOREACH(pe, got_pathlist_head, wanted_refs) {
 		const char *name = pe->path;
 		size_t name_len = pe->path_len;
 
@@ -665,16 +695,13 @@ got_privsep_send_fetch_req(struct imsgbuf *ibuf, int fd,
 			return got_error_from_errno(
 			    "imsg_add FETCH_WANTED_REF");
 
-		wbuf->fd = -1;
 		imsg_close(ibuf, wbuf);
 		err = flush_imsg(ibuf);
 		if (err)
 			return err;
 	}
 
-
 	return NULL;
-
 }
 
 const struct got_error *
@@ -693,6 +720,7 @@ got_privsep_recv_fetch_progress(int *done, struct got_object_id **id,
 	size_t datalen;
 	struct got_imsg_fetch_symrefs *isymrefs = NULL;
 	size_t n, remain;
+	struct got_pathlist_entry *new;
 	off_t off;
 	int i;
 
@@ -755,11 +783,12 @@ got_privsep_recv_fetch_progress(int *done, struct got_object_id **id,
 			}
 			off += s->target_len;
 			remain -= s->target_len;
-			err = got_pathlist_append(symrefs, name, target);
-			if (err) {
+			err = got_pathlist_insert(&new, symrefs, name, target);
+			if (err || new == NULL) {
 				free(name);
 				free(target);
-				goto done;
+				if (err->code != GOT_ERR_REF_DUP_ENTRY)
+					goto done;
 			}
 		}
 		break;
@@ -853,7 +882,6 @@ send_send_ref(const char *name, size_t name_len, struct got_object_id *id,
 	if (imsg_add(wbuf, name, name_len) == -1)
 		return got_error_from_errno("imsg_add SEND_REF");
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 	return flush_imsg(ibuf);
 }
@@ -872,9 +900,9 @@ got_privsep_send_send_req(struct imsgbuf *ibuf, int fd,
 	memset(&zero_id, 0, sizeof(zero_id));
 	memset(&sendreq, 0, sizeof(sendreq));
 	sendreq.verbosity = verbosity;
-	TAILQ_FOREACH(pe, have_refs, entry)
+	RB_FOREACH(pe, got_pathlist_head, have_refs)
 		sendreq.nrefs++;
-	TAILQ_FOREACH(pe, delete_refs, entry)
+	RB_FOREACH(pe, got_pathlist_head, delete_refs)
 		sendreq.nrefs++;
 	if (imsg_compose(ibuf, GOT_IMSG_SEND_REQUEST, 0, 0, fd,
 	    &sendreq, sizeof(sendreq)) == -1) {
@@ -883,12 +911,12 @@ got_privsep_send_send_req(struct imsgbuf *ibuf, int fd,
 		goto done;
 	}
 
+	fd = -1;
 	err = flush_imsg(ibuf);
 	if (err)
 		goto done;
-	fd = -1;
 
-	TAILQ_FOREACH(pe, have_refs, entry) {
+	RB_FOREACH(pe, got_pathlist_head ,have_refs) {
 		const char *name = pe->path;
 		size_t name_len = pe->path_len;
 		struct got_object_id *id = pe->data;
@@ -897,7 +925,7 @@ got_privsep_send_send_req(struct imsgbuf *ibuf, int fd,
 			goto done;
 	}
 
-	TAILQ_FOREACH(pe, delete_refs, entry) {
+	RB_FOREACH(pe, got_pathlist_head, delete_refs) {
 		const char *name = pe->path;
 		size_t name_len = pe->path_len;
 		err = send_send_ref(name, name_len, &zero_id, 1, ibuf);
@@ -908,7 +936,6 @@ done:
 	if (fd != -1 && close(fd) == -1 && err == NULL)
 		err = got_error_from_errno("close");
 	return err;
-
 }
 
 const struct got_error *
@@ -918,13 +945,12 @@ got_privsep_recv_send_remote_refs(struct got_pathlist_head *remote_refs,
 	const struct got_error *err = NULL;
 	struct imsg imsg;
 	size_t datalen;
-	int done = 0;
 	struct got_imsg_send_remote_ref iremote_ref;
 	struct got_object_id *id = NULL;
 	char *refname = NULL;
 	struct got_pathlist_entry *new;
 
-	while (!done) {
+	while (1) {
 		err = got_privsep_recv_imsg(&imsg, ibuf, 0);
 		if (err)
 			return err;
@@ -963,6 +989,7 @@ got_privsep_recv_send_remote_refs(struct got_pathlist_head *remote_refs,
 			}
 			id = NULL;
 			refname = NULL;
+			imsg_free(&imsg);
 			break;
 		case GOT_IMSG_SEND_PACK_REQUEST:
 			if (datalen != 0) {
@@ -970,11 +997,10 @@ got_privsep_recv_send_remote_refs(struct got_pathlist_head *remote_refs,
 				goto done;
 			}
 			/* got-send-pack is now waiting for a pack file. */
-			done = 1;
-			break;
+			goto done;
 		default:
 			err = got_error(GOT_ERR_PRIVSEP_MSG);
-			break;
+			goto done;
 		}
 	}
 done:
@@ -1054,14 +1080,14 @@ got_privsep_recv_send_progress(int *done, off_t *bytes_sent,
 }
 
 const struct got_error *
-got_privsep_send_index_pack_req(struct imsgbuf *ibuf, uint8_t *pack_sha1,
+got_privsep_send_index_pack_req(struct imsgbuf *ibuf, struct got_object_id *id,
     int fd)
 {
 	const struct got_error *err = NULL;
 
 	/* Keep in sync with struct got_imsg_index_pack_request */
 	if (imsg_compose(ibuf, GOT_IMSG_IDXPACK_REQUEST, 0, 0, fd,
-	    pack_sha1, SHA1_DIGEST_LENGTH) == -1) {
+	    id, sizeof(*id)) == -1) {
 		err = got_error_from_errno("imsg_compose INDEX_REQUEST");
 		close(fd);
 		return err;
@@ -1165,8 +1191,7 @@ got_privsep_recv_obj(struct got_object **obj, struct imsgbuf *ibuf)
 {
 	const struct got_error *err = NULL;
 	struct imsg imsg;
-	const size_t min_datalen =
-	    MIN(sizeof(struct got_imsg_error), sizeof(struct got_imsg_object));
+	const size_t min_datalen = sizeof(struct got_imsg_object);
 
 	*obj = NULL;
 
@@ -1387,9 +1412,7 @@ got_privsep_recv_commit(struct got_commit_object **commit, struct imsgbuf *ibuf)
 	const struct got_error *err = NULL;
 	struct imsg imsg;
 	size_t datalen;
-	const size_t min_datalen =
-	    MIN(sizeof(struct got_imsg_error),
-	    sizeof(struct got_imsg_commit_object));
+	const size_t min_datalen = sizeof(struct got_imsg_commit_object);
 
 	*commit = NULL;
 
@@ -1432,10 +1455,16 @@ send_tree_entries_batch(struct imsgbuf *ibuf,
 		return got_error_from_errno("imsg_add TREE_ENTRY");
 
 	for (i = idx0; i <= idxN; i++) {
+		static const char gap[12]; /* for sha1 inside sha256 align */
 		struct got_parsed_tree_entry *pte = &entries[i];
 
 		/* Keep in sync with struct got_imsg_tree_entry definition! */
-		if (imsg_add(wbuf, pte->id, SHA1_DIGEST_LENGTH) == -1)
+		if (imsg_add(wbuf, pte->id, pte->digest_len) == -1)
+			return got_error_from_errno("imsg_add TREE_ENTRY");
+		if (pte->algo == GOT_HASH_SHA1 &&
+		    imsg_add(wbuf, gap, sizeof(gap)) == -1)
+			return got_error_from_errno("imsg_add TREE_ENTRY");
+		if (imsg_add(wbuf, &pte->algo, sizeof(pte->algo)) == -1)
 			return got_error_from_errno("imsg_add TREE_ENTRY");
 		if (imsg_add(wbuf, &pte->mode, sizeof(pte->mode)) == -1)
 			return got_error_from_errno("imsg_add TREE_ENTRY");
@@ -1447,7 +1476,6 @@ send_tree_entries_batch(struct imsgbuf *ibuf,
 			return got_error_from_errno("imsg_add TREE_ENTRY");
 	}
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 	return NULL;
 }
@@ -1560,7 +1588,9 @@ recv_tree_entries(void *data, size_t datalen, struct got_tree_object *tree,
 		te_name = buf + sizeof(ite);
 		memcpy(te->name, te_name, ite.namelen);
 		te->name[ite.namelen] = '\0';
-		memcpy(te->id.sha1, ite.id, SHA1_DIGEST_LENGTH);
+		memcpy(te->id.hash, ite.id, sizeof(te->id.hash));
+		memcpy(te->id.hash, ite.id, got_hash_digest_length(ite.algo));
+		te->id.algo = ite.algo;
 		te->mode = ite.mode;
 		te->idx = *nentries;
 		(*nentries)++;
@@ -1575,9 +1605,7 @@ const struct got_error *
 got_privsep_recv_tree(struct got_tree_object **tree, struct imsgbuf *ibuf)
 {
 	const struct got_error *err = NULL;
-	const size_t min_datalen =
-	    MIN(sizeof(struct got_imsg_error),
-	    sizeof(struct got_imsg_tree_object));
+	const size_t min_datalen = sizeof(struct got_imsg_tree_object);
 	struct got_imsg_tree_object *itree;
 	int nentries = 0;
 
@@ -1832,9 +1860,7 @@ got_privsep_recv_tag(struct got_tag_object **tag, struct imsgbuf *ibuf)
 	struct imsg imsg;
 	struct got_imsg_tag_object *itag;
 	size_t len, datalen;
-	const size_t min_datalen =
-	    MIN(sizeof(struct got_imsg_error),
-	    sizeof(struct got_imsg_tag_object));
+	const size_t min_datalen = sizeof(struct got_imsg_tag_object);
 
 	*tag = NULL;
 
@@ -1945,6 +1971,7 @@ got_privsep_init_pack_child(struct imsgbuf *ibuf, struct got_pack *pack,
 
 	ipackidx.len = packidx->len;
 	ipackidx.packfile_size = pack->filesize;
+	ipackidx.algo = packidx->algo;
 	fd = dup(packidx->fd);
 	if (fd == -1)
 		return got_error_from_errno("dup");
@@ -1960,6 +1987,7 @@ got_privsep_init_pack_child(struct imsgbuf *ibuf, struct got_pack *pack,
 	    sizeof(ipack.path_packfile)) >= sizeof(ipack.path_packfile))
 		return got_error(GOT_ERR_NO_SPACE);
 	ipack.filesize = pack->filesize;
+	ipack.algo = pack->algo;
 
 	fd = dup(pack->fd);
 	if (fd == -1)
@@ -2189,8 +2217,7 @@ got_privsep_recv_gitconfig_int(int *val, struct imsgbuf *ibuf)
 	const struct got_error *err = NULL;
 	struct imsg imsg;
 	size_t datalen;
-	const size_t min_datalen =
-	    MIN(sizeof(struct got_imsg_error), sizeof(int));
+	const size_t min_datalen = sizeof(int);
 
 	*val = 0;
 
@@ -2257,8 +2284,8 @@ got_privsep_recv_gitconfig_remotes(struct got_remote_repo **remotes,
 	switch (imsg.hdr.type) {
 	case GOT_IMSG_GITCONFIG_REMOTES:
 		if (datalen != sizeof(iremotes)) {
-			err = got_error(GOT_ERR_PRIVSEP_LEN);
-			break;
+			imsg_free(&imsg);
+			return got_error(GOT_ERR_PRIVSEP_LEN);
 		}
 		memcpy(&iremotes, imsg.data, sizeof(iremotes));
 		if (iremotes.nremotes == 0) {
@@ -2467,8 +2494,7 @@ got_privsep_recv_gotconfig_remotes(struct got_remote_repo **remotes,
 	size_t datalen;
 	struct got_imsg_remotes iremotes;
 	struct got_imsg_remote iremote;
-	const size_t min_datalen =
-	    MIN(sizeof(struct got_imsg_error), sizeof(iremotes));
+	const size_t min_datalen = sizeof(iremotes);
 
 	*remotes = NULL;
 	*nremotes = 0;
@@ -2508,8 +2534,7 @@ got_privsep_recv_gotconfig_remotes(struct got_remote_repo **remotes,
 
 	while (*nremotes < iremotes.nremotes) {
 		struct got_remote_repo *remote;
-		const size_t min_datalen =
-		    MIN(sizeof(struct got_imsg_error), sizeof(iremote));
+		const size_t min_datalen = sizeof(iremote);
 		int i;
 
 		err = got_privsep_recv_imsg(&imsg, ibuf, min_datalen);
@@ -2674,15 +2699,12 @@ got_privsep_send_commit_traversal_request(struct imsgbuf *ibuf,
 		return got_error_from_errno("imsg_add "
 		    "COMMIT_TRAVERSAL_REQUEST");
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
-
 	return flush_imsg(ibuf);
 }
 
 const struct got_error *
 got_privsep_recv_traversed_commits(struct got_commit_object **changed_commit,
-    struct got_object_id **changed_commit_id,
     struct got_object_id_queue *commit_ids, struct imsgbuf *ibuf)
 {
 	const struct got_error *err = NULL;
@@ -2693,7 +2715,6 @@ got_privsep_recv_traversed_commits(struct got_commit_object **changed_commit,
 	int i, done = 0;
 
 	*changed_commit = NULL;
-	*changed_commit_id = NULL;
 
 	while (!done) {
 		err = got_privsep_recv_imsg(&imsg, ibuf, 0);
@@ -2719,23 +2740,9 @@ got_privsep_recv_traversed_commits(struct got_commit_object **changed_commit,
 				memcpy(&qid->id, &ids[i], sizeof(ids[i]));
 				STAILQ_INSERT_TAIL(commit_ids, qid, entry);
 
-				/* The last commit may contain a change. */
-				if (i == icommits->ncommits - 1) {
-					*changed_commit_id =
-					    got_object_id_dup(&qid->id);
-					if (*changed_commit_id == NULL) {
-						err = got_error_from_errno(
-						    "got_object_id_dup");
-						break;
-					}
-				}
 			}
 			break;
 		case GOT_IMSG_COMMIT:
-			if (*changed_commit_id == NULL) {
-				err = got_error(GOT_ERR_PRIVSEP_MSG);
-				break;
-			}
 			err = get_commit_from_imsg(changed_commit, &imsg,
 			    datalen, ibuf);
 			break;
@@ -2772,14 +2779,13 @@ got_privsep_send_enumerated_tree(size_t *totlen, struct imsgbuf *ibuf,
 	if (wbuf == NULL)
 		return got_error_from_errno("imsg_create ENUMERATED_TREE");
 
-	if (imsg_add(wbuf, tree_id->sha1, SHA1_DIGEST_LENGTH) == -1)
+	if (imsg_add(wbuf, tree_id, sizeof(*tree_id)) == -1)
 		return got_error_from_errno("imsg_add ENUMERATED_TREE");
 	if (imsg_add(wbuf, &nentries, sizeof(nentries)) == -1)
 		return got_error_from_errno("imsg_add ENUMERATED_TREE");
 	if (imsg_add(wbuf, path, path_len) == -1)
 		return got_error_from_errno("imsg_add ENUMERATED_TREE");
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 
 	if (entries) {
@@ -2831,17 +2837,16 @@ got_privsep_send_enumerated_commit(struct imsgbuf *ibuf,
 	struct ibuf *wbuf;
 
 	wbuf = imsg_create(ibuf, GOT_IMSG_ENUMERATED_COMMIT, 0, 0,
-	    sizeof(struct got_imsg_enumerated_commit) + SHA1_DIGEST_LENGTH);
+	    sizeof(struct got_imsg_enumerated_commit));
 	if (wbuf == NULL)
 		return got_error_from_errno("imsg_create ENUMERATED_COMMIT");
 
 	/* Keep in sync with struct got_imsg_enumerated_commit! */
-	if (imsg_add(wbuf, id, SHA1_DIGEST_LENGTH) == -1)
+	if (imsg_add(wbuf, id, sizeof(*id)) == -1)
 		return got_error_from_errno("imsg_add ENUMERATED_COMMIT");
 	if (imsg_add(wbuf, &mtime, sizeof(mtime)) == -1)
 		return got_error_from_errno("imsg_add ENUMERATED_COMMIT");
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 	/* Don't flush yet, tree entries or ENUMERATION_DONE will follow. */
 	return NULL;
@@ -2888,7 +2893,7 @@ got_privsep_recv_enumerated_objects(int *found_all_objects,
 				break;
 			}
 			icommit = (struct got_imsg_enumerated_commit *)imsg.data;
-			memcpy(commit_id.sha1, icommit->id, SHA1_DIGEST_LENGTH);
+			memcpy(&commit_id, &icommit->id, sizeof(commit_id));
 			mtime = icommit->mtime;
 			have_commit = 1;
 			break;
@@ -2908,7 +2913,7 @@ got_privsep_recv_enumerated_objects(int *found_all_objects,
 				err = got_error(GOT_ERR_PRIVSEP_LEN);
 				break;
 			}
-			memcpy(tree_id.sha1, itree->id, sizeof(tree_id.sha1));
+			memcpy(&tree_id, &itree->id, sizeof(tree_id));
 			free(path);
 			path = strndup(imsg.data + sizeof(*itree), path_len);
 			if (path == NULL) {
@@ -3171,7 +3176,6 @@ send_idlist(struct imsgbuf *ibuf, struct got_object_id **ids, size_t nids)
 			return got_error_from_errno("imsg_add OBJ_ID_LIST");
 	}
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 
 	return flush_imsg(ibuf);
@@ -3267,6 +3271,130 @@ got_privsep_recv_object_idlist(int *done, struct got_object_id **ids,
 	return err;
 }
 
+static const struct got_error *
+send_qidlist(struct imsgbuf *ibuf, struct got_object_qid **qids, size_t nids)
+{
+	const struct got_error *err = NULL;
+	struct got_imsg_object_id_queue idqueue;
+	struct ibuf *wbuf;
+	size_t i;
+
+	memset(&idqueue, 0, sizeof(idqueue));
+
+	if (nids > GOT_IMSG_OBJ_ID_QUEUE_MAX_NIDS)
+		return got_error(GOT_ERR_NO_SPACE);
+
+	wbuf = imsg_create(ibuf, GOT_IMSG_OBJ_ID_QUEUE, 0, 0,
+	    sizeof(idqueue) + nids * sizeof(struct got_object_qid));
+	if (wbuf == NULL) {
+		err = got_error_from_errno("imsg_create OBJ_ID_QUEUE");
+		return err;
+	}
+
+	idqueue.nids = nids;
+	if (imsg_add(wbuf, &idqueue, sizeof(idqueue)) == -1)
+		return got_error_from_errno("imsg_add OBJ_ID_QUEUE");
+
+	for (i = 0; i < nids; i++) {
+		struct got_object_qid *qid = qids[i];
+		if (imsg_add(wbuf, qid, sizeof(*qid)) == -1)
+			return got_error_from_errno("imsg_add OBJ_ID_LIST");
+	}
+
+	imsg_close(ibuf, wbuf);
+
+	return flush_imsg(ibuf);
+}
+
+const struct got_error *
+got_privsep_send_object_id_queue(struct imsgbuf *ibuf,
+    struct got_object_id_queue *ids, size_t nids)
+{
+	const struct got_error *err = NULL;
+	struct got_object_qid *qid, *qidlist[GOT_IMSG_OBJ_ID_QUEUE_MAX_NIDS];
+	int i, queued = 0;
+
+	qid = STAILQ_FIRST(ids);
+	for (i = 0; qid && i < nids; i++) {
+		qidlist[i % nitems(qidlist)] = qid;
+		queued++;
+		if (queued >= nitems(qidlist)) {
+			err = send_qidlist(ibuf, qidlist, queued);
+			if (err)
+				return err;
+			queued = 0;
+		}
+		qid = STAILQ_NEXT(qid, entry);
+	}
+
+	if (queued > 0) {
+		err = send_qidlist(ibuf, qidlist, queued);
+		if (err)
+			return err;
+	}
+
+	if (imsg_compose(ibuf, GOT_IMSG_OBJ_ID_QUEUE_DONE, 0, 0, -1, NULL, 0)
+	    == -1)
+		return got_error_from_errno("imsg_compose OBJ_ID_QUEUE_DONE");
+
+	return flush_imsg(ibuf);
+}
+
+const struct got_error *
+got_privsep_recv_object_id_queue(int *done, struct got_object_id_queue *ids,
+    size_t *nids, struct imsgbuf *ibuf)
+{
+	const struct got_error *err = NULL;
+	struct imsg imsg;
+	struct got_imsg_object_id_queue *qidlist;
+	struct got_object_qid *qid, *iqid;
+	size_t datalen;
+
+	*done = 0;
+	*nids = 0;
+
+	err = got_privsep_recv_imsg(&imsg, ibuf, 0);
+	if (err)
+		return err;
+
+	datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
+	switch (imsg.hdr.type) {
+	case GOT_IMSG_OBJ_ID_QUEUE:
+		if (datalen < sizeof(*qidlist)) {
+			err = got_error(GOT_ERR_PRIVSEP_LEN);
+			break;
+		}
+		qidlist = imsg.data;
+		if (qidlist->nids > GOT_IMSG_OBJ_ID_LIST_MAX_NIDS ||
+		    qidlist->nids * sizeof(struct got_object_qid) !=
+		    datalen - sizeof(*qidlist)) {
+			err = got_error(GOT_ERR_PRIVSEP_LEN);
+			break;
+		}
+		iqid = (struct got_object_qid *)(imsg.data + sizeof(*qidlist));
+		while (*nids < qidlist->nids) {
+			err = got_object_qid_alloc_partial(&qid);
+			if (err)
+				break;
+			memcpy(qid, iqid, sizeof(*qid));
+			STAILQ_INSERT_TAIL(ids, qid, entry);
+			(*nids)++;
+			iqid++;
+		}
+		break;
+	case GOT_IMSG_OBJ_ID_QUEUE_DONE:
+		*done = 1;
+		break;
+	default:
+		err = got_error(GOT_ERR_PRIVSEP_MSG);
+		break;
+	}
+
+	imsg_free(&imsg);
+
+	return err;
+}
+
 const struct got_error *
 got_privsep_send_delta_reuse_req(struct imsgbuf *ibuf)
 {
@@ -3308,7 +3436,6 @@ got_privsep_send_reused_deltas(struct imsgbuf *ibuf,
 			return got_error_from_errno("imsg_add REUSED_DELTAS");
 	}
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 
 	return flush_imsg(ibuf);
@@ -3375,8 +3502,7 @@ const struct got_error *
 got_privsep_init_commit_painting(struct imsgbuf *ibuf)
 {
 	if (imsg_compose(ibuf, GOT_IMSG_COMMIT_PAINTING_INIT,
-	    0, 0, -1, NULL, 0)
-	    == -1)
+	    0, 0, -1, NULL, 0) == -1)
 		return got_error_from_errno("imsg_compose "
 		    "COMMIT_PAINTING_INIT");
 
@@ -3384,18 +3510,10 @@ got_privsep_init_commit_painting(struct imsgbuf *ibuf)
 }
 
 const struct got_error *
-got_privsep_send_painting_request(struct imsgbuf *ibuf, int idx,
-    struct got_object_id *id, intptr_t color)
+got_privsep_send_painting_request(struct imsgbuf *ibuf)
 {
-	struct got_imsg_commit_painting_request ireq;
-
-	memset(&ireq, 0, sizeof(ireq));
-	memcpy(&ireq.id, id, sizeof(ireq.id));
-	ireq.idx = idx;
-	ireq.color = color;
-
 	if (imsg_compose(ibuf, GOT_IMSG_COMMIT_PAINTING_REQUEST, 0, 0, -1,
-	    &ireq, sizeof(ireq)) == -1)
+	    NULL, 0) == -1)
 		return got_error_from_errno("imsg_compose "
 		    "COMMIT_PAINTING_REQUEST");
 
@@ -3437,7 +3555,7 @@ send_painted_commits(struct got_object_id_queue *ids, int *nids,
 		color = (intptr_t)qid->data;
 
 		/* Keep in sync with struct got_imsg_painted_commit! */
-		if (imsg_add(wbuf, qid->id.sha1, SHA1_DIGEST_LENGTH) == -1)
+		if (imsg_add(wbuf, &qid->id, sizeof(qid->id)) == -1)
 			return got_error_from_errno("imsg_add PAINTED_COMMITS");
 		if (imsg_add(wbuf, &color, sizeof(color)) == -1)
 			return got_error_from_errno("imsg_add PAINTED_COMMITS");
@@ -3445,9 +3563,7 @@ send_painted_commits(struct got_object_id_queue *ids, int *nids,
 		got_object_qid_free(qid);
 	}
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
-
 	return flush_imsg(ibuf);
 }
 
@@ -3505,18 +3621,26 @@ got_privsep_recv_painted_commits(struct got_object_id_queue *new_ids,
 			return err;
 
 		datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
-		if (imsg.hdr.type == GOT_IMSG_COMMIT_PAINTING_DONE)
-			break;
-		if (imsg.hdr.type != GOT_IMSG_PAINTED_COMMITS)
+		if (imsg.hdr.type == GOT_IMSG_COMMIT_PAINTING_DONE) {
+			imsg_free(&imsg);
+			return NULL;
+		}
+		if (imsg.hdr.type != GOT_IMSG_PAINTED_COMMITS){
+			imsg_free(&imsg);
 			return got_error(GOT_ERR_PRIVSEP_MSG);
+		}
 
-		if (datalen < sizeof(icommits))
+		if (datalen < sizeof(icommits)){
+			imsg_free(&imsg);
 			return got_error(GOT_ERR_PRIVSEP_LEN);
+		}
 		memcpy(&icommits, imsg.data, sizeof(icommits));
 		if (icommits.ncommits * sizeof(icommit) < icommits.ncommits ||
 		    datalen < sizeof(icommits) +
-		    icommits.ncommits * sizeof(icommit))
+		    icommits.ncommits * sizeof(icommit)){
+			imsg_free(&imsg);
 			return got_error(GOT_ERR_PRIVSEP_LEN);
+		}
 
 		for (i = 0; i < icommits.ncommits; i++) {
 			memcpy(&icommit,
@@ -3525,7 +3649,7 @@ got_privsep_recv_painted_commits(struct got_object_id_queue *new_ids,
 
 			if (icommits.present_in_pack) {
 				struct got_object_id id;
-				memcpy(id.sha1, icommit.id, SHA1_DIGEST_LENGTH);
+				memcpy(&id, &icommit.id, sizeof(id));
 				err = cb(cb_arg, &id, icommit.color);
 				if (err)
 					break;
@@ -3534,17 +3658,20 @@ got_privsep_recv_painted_commits(struct got_object_id_queue *new_ids,
 				err = got_object_qid_alloc_partial(&qid);
 				if (err)
 					break;
-				memcpy(qid->id.sha1, icommit.id,
-				    SHA1_DIGEST_LENGTH);
+				memcpy(&qid->id, &icommit.id,
+				    sizeof(qid->id));
 				qid->data = (void *)icommit.color;
-				STAILQ_INSERT_TAIL(new_ids, qid, entry);
+				if (icommit.color == 0 /* COLOR_KEEP */)
+					STAILQ_INSERT_TAIL(new_ids, qid, entry);
+				else
+					STAILQ_INSERT_HEAD(new_ids, qid, entry);
 			}
 		}
 
 		imsg_free(&imsg);
+		if (err)
+			return err;
 	}
-
-	return err;
 }
 
 const struct got_error *
@@ -3563,6 +3690,7 @@ got_privsep_unveil_exec_helpers(void)
 	    GOT_PATH_PROG_FETCH_PACK,
 	    GOT_PATH_PROG_INDEX_PACK,
 	    GOT_PATH_PROG_SEND_PACK,
+	    /* GOT_PATH_PROG_FETCH_HTTP explicitly excluded */
 	};
 	size_t i;
 
