@@ -14,10 +14,9 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-#include "got_compat.h"
-
 #include <sys/types.h>
 #include <sys/queue.h>
+#include <sys/tree.h>
 #include <sys/uio.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
@@ -25,12 +24,15 @@
 #include <sys/resource.h>
 
 #include <ctype.h>
+#include <endian.h>
 #include <fcntl.h>
 #include <fnmatch.h>
 #include <limits.h>
 #include <dirent.h>
 #include <stdlib.h>
 #include <stdio.h>
+#include <sha1.h>
+#include <sha2.h>
 #include <string.h>
 #include <time.h>
 #include <unistd.h>
@@ -38,6 +40,8 @@
 #include <errno.h>
 #include <libgen.h>
 #include <stdint.h>
+#include <imsg.h>
+#include <uuid.h>
 
 #include "bloom.h"
 
@@ -51,13 +55,13 @@
 
 #include "got_lib_delta.h"
 #include "got_lib_delta_cache.h"
+#include "got_lib_hash.h"
 #include "got_lib_inflate.h"
 #include "got_lib_object.h"
 #include "got_lib_object_parse.h"
 #include "got_lib_object_create.h"
 #include "got_lib_pack.h"
 #include "got_lib_privsep.h"
-#include "got_lib_hash.h"
 #include "got_lib_object_cache.h"
 #include "got_lib_repository.h"
 #include "got_lib_gotconfig.h"
@@ -384,6 +388,7 @@ got_repo_temp_fds_set(struct got_repository *repo, int *temp_fds)
 const struct got_error *
 got_repo_temp_fds_get(int *fd, int *idx, struct got_repository *repo)
 {
+	const struct got_error *err = NULL;
 	int i;
 
 	*fd = -1;
@@ -393,8 +398,9 @@ got_repo_temp_fds_get(int *fd, int *idx, struct got_repository *repo)
 		if (repo->tempfile_use_mask & (1 << i))
 			continue;
 		if (repo->tempfiles[i] != -1) {
-			if (ftruncate(repo->tempfiles[i], 0L) == -1)
-				return got_error_from_errno("ftruncate");
+			err = got_opentemp_truncatefd(repo->tempfiles[i]);
+			if (err)
+				return err;
 			*fd = repo->tempfiles[i];
 			*idx = i;
 			repo->tempfile_use_mask |= (1 << i);
@@ -708,7 +714,7 @@ got_repo_open(struct got_repository **repop, const char *path,
 		return got_error_from_errno("calloc");
 
 	RB_INIT(&repo->packidx_bloom_filters);
-	TAILQ_INIT(&repo->packidx_paths);
+	RB_INIT(&repo->packidx_paths);
 
 	for (i = 0; i < nitems(repo->privsep_children); i++) {
 		memset(&repo->privsep_children[i], 0,
@@ -787,7 +793,7 @@ got_repo_open(struct got_repository **repop, const char *path,
 	err = read_gitconfig(repo, global_gitconfig_path);
 	if (err)
 		goto done;
-	if (repo->gitconfig_repository_format_version != 0) {
+	if (repo->gitconfig_repository_format_version > 1) {
 		err = got_error_path(path, GOT_ERR_GIT_REPO_FORMAT);
 		goto done;
 	}
@@ -796,6 +802,18 @@ got_repo_open(struct got_repository **repop, const char *path,
 		char *val = repo->extvals[i];
 		int j, supported = 0;
 
+		if (repo->gitconfig_repository_format_version == 1 &&
+		    strcasecmp(ext, "objectformat") == 0) {
+			if (strcmp(val, "sha1") == 0)
+				continue;
+			if (strcmp(val, "sha256") == 0) {
+				repo->algo = GOT_HASH_SHA256;
+				continue;
+			}
+			err = got_error_path(val, GOT_ERR_OBJECT_FORMAT);
+			goto done;
+		}
+
 		if (!is_boolean_val(val)) {
 			err = got_error_path(ext, GOT_ERR_GIT_REPO_EXT);
 			goto done;
@@ -865,7 +883,7 @@ got_repo_close(struct got_repository *repo)
 	for (i = 0; i < nitems(repo->privsep_children); i++) {
 		if (repo->privsep_children[i].imsg_fd == -1)
 			continue;
-		imsg_clear(repo->privsep_children[i].ibuf);
+		imsgbuf_clear(repo->privsep_children[i].ibuf);
 		free(repo->privsep_children[i].ibuf);
 		err = got_privsep_send_stop(repo->privsep_children[i].imsg_fd);
 		if (err && err->code == GOT_ERR_EOF)
@@ -903,11 +921,117 @@ got_repo_close(struct got_repository *repo)
 	return err;
 }
 
+const struct got_error *
+got_repo_remote_repo_dup(struct got_remote_repo **newp,
+    const struct got_remote_repo *repo)
+{
+	const struct got_error *err = NULL;
+	struct got_remote_repo *new;
+	int i;
+
+	new = calloc(1, sizeof(*new));
+	if (new == NULL)
+		return got_error_from_errno("calloc");
+
+	if (repo->name) {
+		new->name = strdup(repo->name);
+		if (new->name == NULL) {
+			err = got_error_from_errno("strdup");
+			goto done;
+		}
+	}
+
+	if (repo->fetch_url) {
+		new->fetch_url = strdup(repo->fetch_url);
+		if (new->fetch_url == NULL) {
+			err = got_error_from_errno("strdup");
+			goto done;
+		}
+	}
+
+	if (repo->send_url) {
+		new->send_url = strdup(repo->send_url);
+		if (new->send_url == NULL) {
+			err = got_error_from_errno("strdup");
+			goto done;
+		}
+	}
+
+	new->mirror_references = repo->mirror_references;
+
+	new->fetch_all_branches = repo->fetch_all_branches;
+
+	new->nfetch_branches = repo->nfetch_branches;
+	if (repo->fetch_branches) {
+		new->fetch_branches = calloc(repo->nfetch_branches,
+		    sizeof(char *));
+		if (new->fetch_branches == NULL) {
+			err = got_error_from_errno("calloc");
+			goto done;
+		}
+		for (i = 0; i < repo->nfetch_branches; i++) {
+			new->fetch_branches[i] = strdup(
+			    repo->fetch_branches[i]);
+			if (new->fetch_branches[i] == NULL) {
+				err = got_error_from_errno("strdup");
+				goto done;
+			}
+		}
+	}
+
+	new->nsend_branches = repo->nsend_branches;
+	if (repo->send_branches) {
+		new->send_branches = calloc(repo->nsend_branches,
+		    sizeof(char *));
+		if (new->send_branches == NULL) {
+			err = got_error_from_errno("calloc");
+			goto done;
+		}
+		for (i = 0; i < repo->nsend_branches; i++) {
+			new->send_branches[i] = strdup(
+			    repo->send_branches[i]);
+			if (new->send_branches[i] == NULL) {
+				err = got_error_from_errno("strdup");
+				goto done;
+			}
+		}
+	}
+
+	new->nfetch_refs = repo->nfetch_refs;
+	if (repo->fetch_refs) {
+		new->fetch_refs = calloc(repo->nfetch_refs,
+		    sizeof(char *));
+		if (new->fetch_refs == NULL) {
+			err = got_error_from_errno("calloc");
+			goto done;
+		}
+		for (i = 0; i < repo->nfetch_refs; i++) {
+			new->fetch_refs[i] = strdup(
+			    repo->fetch_refs[i]);
+			if (new->fetch_refs[i] == NULL) {
+				err = got_error_from_errno("strdup");
+				goto done;
+			}
+		}
+	}
+done:
+	if (err) {
+		got_repo_free_remote_repo_data(new);
+		free(new);
+	} else
+		*newp = new;
+
+	return err;
+}
+
 void
 got_repo_free_remote_repo_data(struct got_remote_repo *repo)
 {
 	int i;
 
+	if (repo == NULL)
+		return;
+
 	free(repo->name);
 	repo->name = NULL;
 	free(repo->fetch_url);
@@ -924,6 +1048,11 @@ got_repo_free_remote_repo_data(struct got_remote_repo *repo)
 	free(repo->send_branches);
 	repo->send_branches = NULL;
 	repo->nsend_branches = 0;
+	for (i = 0; i < repo->nfetch_refs; i++)
+		free(repo->fetch_refs[i]);
+	free(repo->fetch_refs);
+	repo->fetch_refs = NULL;
+	repo->nfetch_refs = 0;
 }
 
 const struct got_error *
@@ -1067,16 +1196,21 @@ cache_packidx(struct got_repository *repo, struct got_packidx *packidx,
 }
 
 int
-got_repo_is_packidx_filename(const char *name, size_t len)
+got_repo_is_packidx_filename(const char *name, size_t len,
+    enum got_hash_algorithm algo)
 {
-	if (len != GOT_PACKIDX_NAMELEN)
+	size_t digest_string_len;
+
+	digest_string_len = got_hash_digest_string_length(algo);
+
+	if (len != GOT_PACKIDX_NAMELEN(digest_string_len))
 		return 0;
 
 	if (strncmp(name, GOT_PACK_PREFIX, strlen(GOT_PACK_PREFIX)) != 0)
 		return 0;
 
-	if (strcmp(name + strlen(GOT_PACK_PREFIX) +
-	    SHA1_DIGEST_STRING_LENGTH - 1, GOT_PACKIDX_SUFFIX) != 0)
+	if (strcmp(name + strlen(GOT_PACK_PREFIX) + digest_string_len - 1,
+	    GOT_PACKIDX_SUFFIX) != 0)
 		return 0;
 
 	return 1;
@@ -1104,7 +1238,8 @@ got_repo_check_packidx_bloom_filter(struct got_repository *repo,
 
 	bf = get_packidx_bloom_filter(repo, path_packidx, strlen(path_packidx));
 	if (bf)
-		return bloom_check(bf->bloom, id->sha1, sizeof(id->sha1));
+		return bloom_check(bf->bloom, id->hash,
+		    got_hash_digest_length(id->algo));
 
 	/* No bloom filter means this pack index must be searched. */
 	return 1;
@@ -1116,7 +1251,9 @@ add_packidx_bloom_filter(struct got_repository *repo,
 {
 	int i, nobjects = be32toh(packidx->hdr.fanout_table[0xff]);
 	struct got_packidx_bloom_filter *bf;
-	size_t len;
+	size_t len, digest_len;
+
+	digest_len = got_hash_digest_length(repo->algo);
 
 	/*
 	 * Don't use bloom filters for very large pack index files.
@@ -1154,9 +1291,8 @@ add_packidx_bloom_filter(struct got_repository *repo,
 	/* Minimum size supported by our bloom filter is 1000 entries. */
 	bloom_init(bf->bloom, nobjects < 1000 ? 1000 : nobjects, 0.1);
 	for (i = 0; i < nobjects; i++) {
-		struct got_packidx_object_id *id;
-		id = &packidx->hdr.sorted_ids[i];
-		bloom_add(bf->bloom, id->sha1, sizeof(id->sha1));
+		uint8_t *id = packidx->hdr.sorted_ids + i * digest_len;
+		bloom_add(bf->bloom, id, digest_len);
 	}
 
 	RB_INSERT(got_packidx_bloom_filter_tree,
@@ -1169,9 +1305,9 @@ purge_packidx_paths(struct got_pathlist_head *packidx_paths)
 {
 	struct got_pathlist_entry *pe;
 
-	while (!TAILQ_EMPTY(packidx_paths)) {
-		pe = TAILQ_FIRST(packidx_paths);
-		TAILQ_REMOVE(packidx_paths, pe, entry);
+	while (!RB_EMPTY(packidx_paths)) {
+		pe = RB_MIN(got_pathlist_head, packidx_paths);
+		RB_REMOVE(got_pathlist_head, packidx_paths, pe);
 		free((char *)pe->path);
 		free(pe);
 	}
@@ -1193,7 +1329,7 @@ refresh_packidx_paths(struct got_repository *repo)
 			err = got_error_from_errno2("stat", objects_pack_dir);
 			goto done;
 		}
-	} else if (TAILQ_EMPTY(&repo->packidx_paths) ||
+	} else if (RB_EMPTY(&repo->packidx_paths) ||
 	    sb.st_mtim.tv_sec != repo->pack_path_mtime.tv_sec ||
 	    sb.st_mtim.tv_nsec != repo->pack_path_mtime.tv_nsec) {
 		purge_packidx_paths(&repo->packidx_paths);
@@ -1248,7 +1384,7 @@ got_repo_search_packidx(struct got_packidx **packidx, int *idx,
 	if (err)
 		return err;
 
-	TAILQ_FOREACH(pe, &repo->packidx_paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head, &repo->packidx_paths) {
 		const char *path_packidx = pe->path;
 		int is_cached = 0;
 
@@ -1269,9 +1405,14 @@ got_repo_search_packidx(struct got_packidx **packidx, int *idx,
 			continue; /* already searched */
 
 		err = got_packidx_open(packidx, got_repo_get_fd(repo),
-		    path_packidx, 0);
-		if (err)
+		    path_packidx, 0, repo->algo);
+		if (err) {
+			if (err->code == GOT_ERR_LONELY_PACKIDX) {
+				err = NULL;
+				continue;
+			}
 			goto done;
+		}
 
 		err = add_packidx_bloom_filter(repo, *packidx, path_packidx);
 		if (err)
@@ -1315,6 +1456,7 @@ got_repo_list_packidx(struct got_pathlist_head *packidx_paths,
 	packdir = fdopendir(packdir_fd);
 	if (packdir == NULL) {
 		err = got_error_from_errno("fdopendir");
+		close(packdir_fd);
 		goto done;
 	}
 
@@ -1326,8 +1468,8 @@ got_repo_list_packidx(struct got_pathlist_head *packidx_paths,
 	repo->pack_path_mtime.tv_nsec = sb.st_mtim.tv_nsec;
 
 	while ((dent = readdir(packdir)) != NULL) {
-		if (!got_repo_is_packidx_filename(dent->d_name,
-		    strlen(dent->d_name)))
+		if (!got_repo_is_packidx_filename(dent->d_name, dent->d_namlen,
+		    repo->algo))
 			continue;
 
 		if (asprintf(&path_packidx, "%s/%s", GOT_OBJECTS_PACK_DIR,
@@ -1337,7 +1479,7 @@ got_repo_list_packidx(struct got_pathlist_head *packidx_paths,
 			break;
 		}
 
-		err = got_pathlist_append(packidx_paths, path_packidx, NULL);
+		err = got_pathlist_insert(NULL, packidx_paths, path_packidx, NULL);
 		if (err)
 			break;
 	}
@@ -1371,7 +1513,7 @@ got_repo_get_packidx(struct got_packidx **packidx, const char *path_packidx,
 	/* No luck. Search the filesystem. */
 
 	err = got_packidx_open(packidx, got_repo_get_fd(repo),
-	    path_packidx, 0);
+	    path_packidx, 0, repo->algo);
 	if (err)
 		return err;
 
@@ -1439,6 +1581,7 @@ got_repo_cache_pack(struct got_pack **packp, struct got_repository *repo,
 {
 	const struct got_error *err = NULL;
 	struct got_pack *pack = NULL;
+	struct got_pack tmp;
 	struct stat sb;
 	size_t i;
 
@@ -1454,7 +1597,6 @@ got_repo_cache_pack(struct got_pack **packp, struct got_repository *repo,
 	}
 
 	if (i == repo->pack_cache_size) {
-		struct got_pack tmp;
 		do {
 			i--;
 		} while (i > 0 && repo->pinned_pack >= 0 &&
@@ -1462,10 +1604,15 @@ got_repo_cache_pack(struct got_pack **packp, struct got_repository *repo,
 		err = got_pack_close(&repo->packs[i]);
 		if (err)
 			return err;
-		if (ftruncate(repo->packs[i].basefd, 0L) == -1)
-			return got_error_from_errno("ftruncate");
-		if (ftruncate(repo->packs[i].accumfd, 0L) == -1)
-			return got_error_from_errno("ftruncate");
+		err = got_opentemp_truncatefd(repo->packs[i].basefd);
+		if (err)
+			return err;
+		err = got_opentemp_truncatefd(repo->packs[i].accumfd);
+		if (err)
+			return err;
+	}
+
+	if (i != 0) {
 		memcpy(&tmp, &repo->packs[i], sizeof(tmp));
 		memcpy(&repo->packs[i], &repo->packs[0],
 		    sizeof(repo->packs[i]));
@@ -1494,6 +1641,7 @@ got_repo_cache_pack(struct got_pack **packp, struct got_repository *repo,
 		goto done;
 	}
 	pack->filesize = sb.st_size;
+	pack->algo = repo->algo;
 
 	pack->privsep_child = NULL;
 
@@ -1586,68 +1734,6 @@ got_repo_unpin_pack(struct got_repository *repo)
 	repo->pinned_pid = 0;
 }
 
-const struct got_error *
-got_repo_init(const char *repo_path, const char *head_name)
-{
-	const struct got_error *err = NULL;
-	const char *dirnames[] = {
-		GOT_OBJECTS_DIR,
-		GOT_OBJECTS_PACK_DIR,
-		GOT_REFS_DIR,
-	};
-	const char *description_str = "Unnamed repository; "
-	    "edit this file 'description' to name the repository.";
-	const char *headref = "ref: refs/heads/";
-	const char *gitconfig_str = "[core]\n"
-	    "\trepositoryformatversion = 0\n"
-	    "\tfilemode = true\n"
-	    "\tbare = true\n";
-	char *headref_str, *path;
-	size_t i;
-
-	if (!got_path_dir_is_empty(repo_path))
-		return got_error(GOT_ERR_DIR_NOT_EMPTY);
-
-	for (i = 0; i < nitems(dirnames); i++) {
-		if (asprintf(&path, "%s/%s", repo_path, dirnames[i]) == -1) {
-			return got_error_from_errno("asprintf");
-		}
-		err = got_path_mkdir(path);
-		free(path);
-		if (err)
-			return err;
-	}
-
-	if (asprintf(&path, "%s/%s", repo_path, "description") == -1)
-		return got_error_from_errno("asprintf");
-	err = got_path_create_file(path, description_str);
-	free(path);
-	if (err)
-		return err;
-
-	if (asprintf(&path, "%s/%s", repo_path, GOT_HEAD_FILE) == -1)
-		return got_error_from_errno("asprintf");
-	if (asprintf(&headref_str, "%s%s", headref,
-	    head_name ? head_name : "main") == -1) {
-		free(path);
-		return got_error_from_errno("asprintf");
-	}
-	err = got_path_create_file(path, headref_str);
-	free(headref_str);
-	free(path);
-	if (err)
-		return err;
-
-	if (asprintf(&path, "%s/%s", repo_path, "config") == -1)
-		return got_error_from_errno("asprintf");
-	err = got_path_create_file(path, gitconfig_str);
-	free(path);
-	if (err)
-		return err;
-
-	return NULL;
-}
-
 static const struct got_error *
 match_packed_object(struct got_object_id **unique_id,
     struct got_repository *repo, const char *id_str_prefix, int obj_type)
@@ -1655,6 +1741,9 @@ match_packed_object(struct got_object_id **unique_id,
 	const struct got_error *err = NULL;
 	struct got_object_id_queue matched_ids;
 	struct got_pathlist_entry *pe;
+	struct timespec tv;
+	int retries = 0;
+	const int max_retries = 10;
 
 	STAILQ_INIT(&matched_ids);
 
@@ -1662,15 +1751,37 @@ match_packed_object(struct got_object_id **unique_id,
 	if (err)
 		return err;
 
-	TAILQ_FOREACH(pe, &repo->packidx_paths, entry) {
-		const char *path_packidx = pe->path;
+	/*
+	 * Opening objects while iterating over the pack-index path
+	 * list is racy. If the set of pack files in the repository
+	 * changes during loop iteration, refresh_packidx_paths() will
+	 * be called again, via got_object_get_type(), invalidating
+	 * the packidx_paths list we are iterating over.
+	 * To work around this we keep track of the current modification
+	 * time and retry the entire loop if it changes.
+	 */
+retry:
+	tv.tv_sec = repo->pack_path_mtime.tv_sec;
+	tv.tv_nsec = repo->pack_path_mtime.tv_nsec;
+
+	RB_FOREACH(pe, got_pathlist_head, &repo->packidx_paths) {
+		const char *path_packidx;
 		struct got_packidx *packidx;
 		struct got_object_qid *qid;
 
+		path_packidx = pe->path;
+
 		err = got_packidx_open(&packidx, got_repo_get_fd(repo),
-		    path_packidx, 0);
-		if (err)
+		    path_packidx, 0, repo->algo);
+		if (err) {
+			if (err->code == GOT_ERR_LONELY_PACKIDX) {
+				err = NULL;
+				continue;
+			}
 			break;
+		}
+
+		got_object_id_queue_free(&matched_ids);
 
 		err = got_packidx_match_id_str_prefix(&matched_ids,
 		    packidx, id_str_prefix);
@@ -1689,6 +1800,24 @@ match_packed_object(struct got_object_id **unique_id,
 				    &qid->id);
 				if (err)
 					goto done;
+				/*
+				 * If the modification time of the
+				 * 'objects/pack' directory has changed then
+				 * 'pe' could now be an invalid pointer.
+				 */
+				if (tv.tv_sec !=
+				    repo->pack_path_mtime.tv_sec ||
+				    tv.tv_nsec !=
+				    repo->pack_path_mtime.tv_nsec) {
+					if (++retries > max_retries) {
+						err = got_error_msg(
+						    GOT_ERR_TIMEOUT,
+						    "too many concurrent "
+						    "pack file modifications");
+						goto done;
+					}
+					goto retry;
+				}
 				if (matched_type != obj_type)
 					continue;
 			}
@@ -1807,15 +1936,16 @@ got_repo_match_object_id_prefix(struct got_object_id **id,
 {
 	const struct got_error *err = NULL;
 	char *path_objects = NULL, *object_dir = NULL;
-	size_t len;
+	size_t len, digest_string_len;
 	int i;
 
 	*id = NULL;
 
 	path_objects = got_repo_get_path_objects(repo);
+	digest_string_len = got_hash_digest_string_length(repo->algo);
 
 	len = strlen(id_str_prefix);
-	if (len > SHA1_DIGEST_STRING_LENGTH - 1) {
+	if (len > digest_string_len - 1) {
 		err = got_error_path(id_str_prefix, GOT_ERR_BAD_OBJ_ID_STR);
 		goto done;
 	}
@@ -2005,13 +2135,31 @@ got_repo_object_match_tag(struct got_tag_object **tag, const char *name,
 	return err;
 }
 
+const struct got_error *
+got_repo_find_object_id(struct got_object_id *id, struct got_repository *repo)
+{
+	const struct got_error *err;
+	struct got_object_id *matched_id = NULL;
+	char *id_str = NULL;
+
+	err = got_object_id_str(&id_str, id);
+	if (err)
+		return err;
+
+	err = got_repo_match_object_id_prefix(&matched_id, id_str,
+	    GOT_OBJ_TYPE_ANY, repo);
+	free(matched_id);
+	free(id_str);
+	return err;
+}
+
 static const struct got_error *
 alloc_added_blob_tree_entry(struct got_tree_entry **new_te,
     const char *name, mode_t mode, struct got_object_id *blob_id)
 {
 	const struct got_error *err = NULL;
 
-	 *new_te = NULL;
+	*new_te = NULL;
 
 	*new_te = calloc(1, sizeof(**new_te));
 	if (*new_te == NULL)
@@ -2069,6 +2217,41 @@ done:
 	return err;
 }
 
+static const struct got_error *
+insert_dir_entry(struct dirent *de, long pos,
+    struct got_pathlist_head *dirents)
+{
+	const struct got_error *err = NULL;
+	struct got_pathlist_entry *new_pe;
+	char *d_name;
+	long *ppos;
+
+	d_name = strdup(de->d_name);
+	if (d_name == NULL)
+		return got_error_from_errno("strdup");
+
+	/* Record the offset to this dirent in struct DIR as path data. */
+	ppos = malloc(sizeof(*ppos));
+	if (ppos == NULL) {
+		err = got_error_from_errno("malloc");
+		goto done;
+	}
+	*ppos = pos;
+
+	err = got_pathlist_insert(&new_pe, dirents, d_name, ppos);
+	if (err)
+		goto done;
+
+	if (new_pe == NULL)
+		err = got_error(GOT_ERR_TREE_DUP_ENTRY);
+done:
+	if (err) {
+		free(d_name);
+		free(ppos);
+	}
+	return err;
+}
+
 static const struct got_error *
 insert_tree_entry(struct got_tree_entry *new_te,
     struct got_pathlist_head *paths)
@@ -2127,6 +2310,37 @@ done:
 	return err;
 }
 
+static const struct got_error *
+match_ignores(int *ignore, char *d_name, int type,
+    struct got_pathlist_head *ignores)
+{
+	struct got_pathlist_entry *pe;
+
+	*ignore = 0;
+
+	RB_FOREACH(pe, got_pathlist_head, ignores) {
+		if (type == DT_DIR && pe->path_len > 0 &&
+		    pe->path[pe->path_len - 1] == '/') {
+			char stripped[PATH_MAX];
+
+			if (strlcpy(stripped, pe->path,
+			    sizeof(stripped)) >= sizeof(stripped))
+				return got_error(GOT_ERR_NO_SPACE);
+
+			got_path_strip_trailing_slashes(stripped);
+			if (fnmatch(stripped, d_name, 0) == 0) {
+				*ignore = 1;
+				break;
+			}
+		} else if (fnmatch(pe->path, d_name, 0) == 0) {
+			*ignore = 1;
+			break;
+		}
+	}
+
+	return NULL;
+}
+
 static const struct got_error *
 write_tree(struct got_object_id **new_tree_id, const char *path_dir,
     struct got_pathlist_head *ignores, struct got_repository *repo,
@@ -2137,12 +2351,14 @@ write_tree(struct got_object_id **new_tree_id, const char *path_dir,
 	struct dirent *de;
 	int nentries;
 	struct got_tree_entry *new_te = NULL;
-	struct got_pathlist_head paths;
+	struct got_pathlist_head paths, dirents;
 	struct got_pathlist_entry *pe;
+	long pos;
 
 	*new_tree_id = NULL;
 
-	TAILQ_INIT(&paths);
+	RB_INIT(&paths);
+	RB_INIT(&dirents);
 
 	dir = opendir(path_dir);
 	if (dir == NULL) {
@@ -2150,39 +2366,53 @@ write_tree(struct got_object_id **new_tree_id, const char *path_dir,
 		goto done;
 	}
 
-	nentries = 0;
+	/*
+	 * Sort dirents into a pathlist. Otherwise subdir-iteration order
+	 * depends on the readdir() function, which varies between different
+	 * operating systems, making regression testing more difficult.
+	 * This smells of TOCTOU, but if dirents appear, disappear, or change
+	 * during import then we have a racy situation no matter what.
+	 */
+	pos = telldir(dir);
+	if (pos == -1) {
+		err = got_error_from_errno2("telldir", path_dir);
+		goto done;
+	}
 	while ((de = readdir(dir)) != NULL) {
+		if (strcmp(de->d_name, ".") != 0 &&
+		    strcmp(de->d_name, "..") != 0) {
+			err = insert_dir_entry(de, pos, &dirents);
+			if (err)
+				goto done;
+		}
+		pos = telldir(dir);
+		if (pos == -1) {
+			err = got_error_from_errno2("telldir", path_dir);
+			goto done;
+		}
+	}
+
+	nentries = 0;
+	RB_FOREACH(pe, got_pathlist_head, &dirents) {
 		int ignore = 0;
 		int type;
+		long *pos = pe->data;
 
-		if (strcmp(de->d_name, ".") == 0 ||
-		    strcmp(de->d_name, "..") == 0)
-			continue;
+		seekdir(dir, *pos);
+		de = readdir(dir);
+		if (de == NULL) { /* should not happen */
+			err = got_error_fmt(GOT_ERR_EOF,
+			    "unexpected EOF on directory '%s' while seeking "
+			    "to entry '%s' at offset '%ld'\n", path_dir,
+			    pe->path, *pos);
+			goto done;
+		}
 
 		err = got_path_dirent_type(&type, path_dir, de);
 		if (err)
 			goto done;
 
-		TAILQ_FOREACH(pe, ignores, entry) {
-			if (type == DT_DIR && pe->path_len > 0 &&
-			    pe->path[pe->path_len - 1] == '/') {
-				char stripped[PATH_MAX];
-
-				if (strlcpy(stripped, pe->path,
-				    sizeof(stripped)) >= sizeof(stripped)) {
-					err = got_error(GOT_ERR_NO_SPACE);
-					goto done;
-				}
-				got_path_strip_trailing_slashes(stripped);
-				if (fnmatch(stripped, de->d_name, 0) == 0) {
-					ignore = 1;
-					break;
-				}
-			} else if (fnmatch(pe->path, de->d_name, 0) == 0) {
-				ignore = 1;
-				break;
-			}
-		}
+		err = match_ignores(&ignore, de->d_name, type, ignores);
 		if (ignore)
 			continue;
 
@@ -2208,13 +2438,13 @@ write_tree(struct got_object_id **new_tree_id, const char *path_dir,
 		nentries++;
 	}
 
-	if (TAILQ_EMPTY(&paths)) {
+	if (RB_EMPTY(&paths)) {
 		err = got_error_msg(GOT_ERR_NO_TREE_ENTRY,
 		    "cannot create tree without any entries");
 		goto done;
 	}
 
-	TAILQ_FOREACH(pe, &paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head,  &paths) {
 		struct got_tree_entry *te = pe->data;
 		char *path;
 		if (!S_ISREG(te->mode) && !S_ISLNK(te->mode))
@@ -2233,6 +2463,7 @@ write_tree(struct got_object_id **new_tree_id, const char *path_dir,
 done:
 	if (dir)
 		closedir(dir);
+	got_pathlist_free(&dirents, GOT_PATHLIST_FREE_ALL);
 	got_pathlist_free(&paths, GOT_PATHLIST_FREE_NONE);
 	return err;
 }
@@ -2380,12 +2611,13 @@ got_repo_get_packfile_info(int *npackfiles, int *nobjects,
 	packdir = fdopendir(packdir_fd);
 	if (packdir == NULL) {
 		err = got_error_from_errno("fdopendir");
+		close(packdir_fd);
 		goto done;
 	}
 
 	while ((dent = readdir(packdir)) != NULL) {
-		if (!got_repo_is_packidx_filename(dent->d_name,
-		    strlen(dent->d_name)))
+		if (!got_repo_is_packidx_filename(dent->d_name, dent->d_namlen,
+		    repo->algo))
 			continue;
 
 		if (asprintf(&path_packidx, "%s/%s", GOT_OBJECTS_PACK_DIR,
@@ -2395,10 +2627,15 @@ got_repo_get_packfile_info(int *npackfiles, int *nobjects,
 		}
 
 		err = got_packidx_open(&packidx, got_repo_get_fd(repo),
-		    path_packidx, 0);
+		    path_packidx, 0, repo->algo);
 		free(path_packidx);
-		if (err)
+		if (err) {
+			if (err->code == GOT_ERR_LONELY_PACKIDX) {
+				err = NULL;
+				continue;
+			}
 			goto done;
+		}
 
 		if (fstat(packidx->fd, &sb) == -1)
 			goto done;
