@@ -14,10 +14,9 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-#include "got_compat.h"
-
 #include <sys/stat.h>
 #include <sys/queue.h>
+#include <sys/tree.h>
 
 #include <dirent.h>
 #include <limits.h>
@@ -29,9 +28,13 @@
 #include <fcntl.h>
 #include <errno.h>
 #include <unistd.h>
+#include <sha1.h>
+#include <sha2.h>
 #include <zlib.h>
 #include <fnmatch.h>
 #include <libgen.h>
+#include <uuid.h>
+#include <util.h>
 
 #include "got_error.h"
 #include "got_repository.h"
@@ -560,13 +563,13 @@ get_fileindex_path(char **fileindex_path, struct got_worktree *worktree)
 
 static const struct got_error *
 open_fileindex(struct got_fileindex **fileindex, char **fileindex_path,
-    struct got_worktree *worktree)
+    struct got_worktree *worktree, enum got_hash_algorithm algo)
 {
 	const struct got_error *err = NULL;
 	FILE *index = NULL;
 
 	*fileindex_path = NULL;
-	*fileindex = got_fileindex_alloc();
+	*fileindex = got_fileindex_alloc(algo);
 	if (*fileindex == NULL)
 		return got_error_from_errno("got_fileindex_alloc");
 
@@ -579,7 +582,7 @@ open_fileindex(struct got_fileindex **fileindex, char **fileindex_path,
 		if (errno != ENOENT)
 			err = got_error_from_errno2("fopen", *fileindex_path);
 	} else {
-		err = got_fileindex_read(*fileindex, index);
+		err = got_fileindex_read(*fileindex, index, algo);
 		if (fclose(index) == EOF && err == NULL)
 			err = got_error_from_errno("fclose");
 	}
@@ -692,8 +695,11 @@ status_old_new(void *arg, struct got_fileindex_entry *ie,
 	struct diff_dir_cb_arg *a = arg;
 	char *abspath;
 
-	if (a->cancel_cb && a->cancel_cb(a->cancel_arg))
-		return got_error(GOT_ERR_CANCELLED);
+	if (a->cancel_cb) {
+		err = a->cancel_cb(a->cancel_arg);
+		if (err)
+			return err;
+	}
 
 	if (got_path_cmp(parent_path, a->status_path,
 	    strlen(parent_path), a->status_path_len) != 0 &&
@@ -719,12 +725,16 @@ status_old_new(void *arg, struct got_fileindex_entry *ie,
 static const struct got_error *
 status_old(void *arg, struct got_fileindex_entry *ie, const char *parent_path)
 {
+	const struct got_error *err = NULL;
 	struct diff_dir_cb_arg *a = arg;
 	struct got_object_id blob_id, commit_id;
 	unsigned char status;
 
-	if (a->cancel_cb && a->cancel_cb(a->cancel_arg))
-		return got_error(GOT_ERR_CANCELLED);
+	if (a->cancel_cb) {
+		err = a->cancel_cb(a->cancel_arg);
+		if (err)
+			return err;
+	}
 
 	if (!got_path_is_child(ie->path, a->status_path, a->status_path_len))
 		return NULL;
@@ -744,7 +754,7 @@ free_ignores(struct got_pathlist_head *ignores)
 {
 	struct got_pathlist_entry *pe;
 
-	TAILQ_FOREACH(pe, ignores, entry) {
+	RB_FOREACH(pe, got_pathlist_head, ignores) {
 		struct got_pathlist_head *ignorelist = pe->data;
 
 		got_pathlist_free(ignorelist, GOT_PATHLIST_FREE_PATH);
@@ -765,7 +775,7 @@ read_ignores(struct got_pathlist_head *ignores, const char *path, FILE *f)
 	ignorelist = calloc(1, sizeof(*ignorelist));
 	if (ignorelist == NULL)
 		return got_error_from_errno("calloc");
-	TAILQ_INIT(ignorelist);
+	RB_INIT(ignorelist);
 
 	while ((linelen = getline(&line, &linesize, f)) != -1) {
 		if (linelen > 0 && line[linelen - 1] == '\n')
@@ -834,11 +844,11 @@ match_ignores(struct got_pathlist_head *ignores, const char *path)
 	struct got_pathlist_entry *pe;
 
 	/* Handle patterns which match in all directories. */
-	TAILQ_FOREACH(pe, ignores, entry) {
+	RB_FOREACH(pe, got_pathlist_head, ignores) {
 		struct got_pathlist_head *ignorelist = pe->data;
 		struct got_pathlist_entry *pi;
 
-		TAILQ_FOREACH(pi, ignorelist, entry) {
+		RB_FOREACH(pi, got_pathlist_head, ignorelist) {
 			const char *p;
 
 			if (pi->path_len < 3 ||
@@ -866,12 +876,12 @@ match_ignores(struct got_pathlist_head *ignores, const char *path)
 	 * parents, so we can find the most specific ignorelist by walking
 	 * ignores backwards.
 	 */
-	pe = TAILQ_LAST(ignores, got_pathlist_head);
+	pe = RB_MAX(got_pathlist_head, ignores);
 	while (pe) {
 		if (got_path_is_child(path, pe->path, pe->path_len)) {
 			struct got_pathlist_head *ignorelist = pe->data;
 			struct got_pathlist_entry *pi;
-			TAILQ_FOREACH(pi, ignorelist, entry) {
+			RB_FOREACH(pi, got_pathlist_head, ignorelist) {
 				int flags = FNM_LEADING_DIR;
 				if (strstr(pi->path, "/**/") == NULL)
 					flags |= FNM_PATHNAME;
@@ -881,7 +891,7 @@ match_ignores(struct got_pathlist_head *ignores, const char *path)
 				return 1;
 			}
 		}
-		pe = TAILQ_PREV(pe, got_pathlist_head, entry);
+		pe = RB_PREV(got_pathlist_head, ignores,  pe);
 	}
 
 	return 0;
@@ -946,8 +956,11 @@ status_new(int *ignore, void *arg, struct dirent *de, const char *parent_path,
 	if (ignore != NULL)
 		*ignore = 0;
 
-	if (a->cancel_cb && a->cancel_cb(a->cancel_arg))
-		return got_error(GOT_ERR_CANCELLED);
+	if (a->cancel_cb) {
+		err = a->cancel_cb(a->cancel_arg);
+		if (err)
+			return err;
+	}
 
 	if (parent_path[0]) {
 		if (asprintf(&path, "%s/%s", parent_path, de->d_name) == -1)
@@ -1091,7 +1104,7 @@ find_missing_children(void *arg, struct got_fileindex_entry *ie)
 	}
 
 	if (got_path_is_child(ie->path, a->parent_path, a->parent_len))
-		err = got_pathlist_append(a->children, ie->path, NULL);
+		err = got_pathlist_insert(NULL, a->children, ie->path, NULL);
 
 	return err;
 }
@@ -1108,7 +1121,7 @@ report_children(struct got_pathlist_head *children,
 	struct got_pathlist_entry *pe;
 	char *ondisk_path = NULL;
 
-	TAILQ_FOREACH(pe, children, entry) {
+	RB_FOREACH(pe, got_pathlist_head, children) {
 		if (cancel_cb) {
 			err = cancel_cb(cancel_arg);
 			if (err)
@@ -1151,8 +1164,8 @@ worktree_status(struct got_worktree *worktree, const char *path,
 	struct got_pathlist_head ignores, missing_children;
 	struct got_fileindex_entry *ie;
 
-	TAILQ_INIT(&ignores);
-	TAILQ_INIT(&missing_children);
+	RB_INIT(&ignores);
+	RB_INIT(&missing_children);
 
 	if (asprintf(&ondisk_path, "%s%s%s",
 	    worktree->root_path, path[0] ? "/" : "", path) == -1)
@@ -1189,7 +1202,7 @@ worktree_status(struct got_worktree *worktree, const char *path,
 				if (err)
 					goto done;
 			}
-			if (TAILQ_EMPTY(&missing_children)) {
+			if (RB_EMPTY(&missing_children)) {
 				err = report_single_file_status(path,
 				    ondisk_path, fileindex,
 				    status_cb, status_arg, repo,
@@ -1461,7 +1474,7 @@ append_ct_diff(struct got_commitable *ct, int *diff_header_shown,
 			goto done;
 	}
 
-	err = got_diff_blob_file(blob1, f1, size1, label1,
+	err = got_diff_blob_file(NULL, NULL, blob1, f1, size1, label1,
 	    ondisk_file ? ondisk_file : f2, f2_exists, &sb, ct->path,
 	    GOT_DIFF_ALGORITHM_PATIENCE, 3, 0, 0, NULL, diff_outfile);
 done:
@@ -1825,7 +1838,7 @@ match_modified_subtree(int *modified, struct got_tree_entry *te,
 	    te->name) == -1)
 		return got_error_from_errno("asprintf");
 
-	TAILQ_FOREACH(pe, commitable_paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head, commitable_paths) {
 		struct got_commitable *ct = pe->data;
 		*modified = got_path_is_child(ct->in_repo_path, te_path,
 		    strlen(te_path));
@@ -1847,7 +1860,7 @@ match_deleted_or_modified_ct(struct got_commitable **ctp,
 
 	*ctp = NULL;
 
-	TAILQ_FOREACH(pe, commitable_paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head, commitable_paths) {
 		struct got_commitable *ct = pe->data;
 		char *ct_name = NULL;
 		int path_matches;
@@ -1947,11 +1960,11 @@ write_tree(struct got_object_id **new_tree_id, int *nentries,
 	struct got_tree_entry *te, *new_te = NULL;
 	struct got_pathlist_entry *pe;
 
-	TAILQ_INIT(&paths);
+	RB_INIT(&paths);
 	*nentries = 0;
 
 	/* Insert, and recurse into, newly added entries first. */
-	TAILQ_FOREACH(pe, commitable_paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head, commitable_paths) {
 		struct got_commitable *ct = pe->data;
 		char *child_path = NULL, *slash;
 
@@ -2109,7 +2122,7 @@ update_fileindex_after_commit(struct got_worktree *worktree,
 	struct got_pathlist_entry *pe;
 	char *relpath = NULL;
 
-	TAILQ_FOREACH(pe, commitable_paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head, commitable_paths) {
 		struct got_fileindex_entry *ie;
 		struct got_commitable *ct = pe->data;
 
@@ -2132,22 +2145,20 @@ update_fileindex_after_commit(struct got_worktree *worktree,
 
 				err = got_fileindex_entry_update(ie,
 				    worktree->root_fd, relpath,
-				    ct->staged_blob_id->sha1,
-				    new_base_commit_id->sha1,
+				    ct->staged_blob_id, new_base_commit_id,
 				    !have_staged_files);
 			} else
 				err = got_fileindex_entry_update(ie,
 				    worktree->root_fd, relpath,
-				    ct->blob_id->sha1,
-				    new_base_commit_id->sha1,
+				    ct->blob_id, new_base_commit_id,
 				    !have_staged_files);
 		} else {
 			err = got_fileindex_entry_alloc(&ie, pe->path);
 			if (err)
 				goto done;
 			err = got_fileindex_entry_update(ie,
-			    worktree->root_fd, relpath, ct->blob_id->sha1,
-			    new_base_commit_id->sha1, 1);
+			    worktree->root_fd, relpath, ct->blob_id,
+			    new_base_commit_id, 1);
 			if (err) {
 				got_fileindex_entry_free(ie);
 				goto done;
@@ -2218,8 +2229,8 @@ commit_worktree(struct got_object_id **new_commit_id,
     struct got_object_id *head_commit_id,
     struct got_object_id *parent_id2,
     struct got_worktree *worktree,
-    const char *author, const char *committer, char *diff_path,
-    got_worktree_commit_msg_cb commit_msg_cb, void *commit_arg,
+    const char *author, time_t author_time, const char *committer,
+    char *diff_path, got_worktree_commit_msg_cb commit_msg_cb, void *commit_arg,
     got_worktree_status_cb status_cb, void *status_arg,
     struct got_repository *repo)
 {
@@ -2259,7 +2270,7 @@ commit_worktree(struct got_object_id **new_commit_id,
 	}
 
 	/* Create blobs from added and modified files and record their IDs. */
-	TAILQ_FOREACH(pe, commitable_paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head, commitable_paths) {
 		struct got_commitable *ct = pe->data;
 		char *ondisk_path;
 
@@ -2304,8 +2315,10 @@ commit_worktree(struct got_object_id **new_commit_id,
 		nparents++;
 	}
 	timestamp = time(NULL);
+	if (committer == NULL)
+		author_time = timestamp;
 	err = got_object_commit_create(new_commit_id, new_tree_id, &parent_ids,
-	    nparents, author, timestamp, committer, timestamp, logmsg, repo);
+	    nparents, author, author_time, committer, timestamp, logmsg, repo);
 	if (logmsg != NULL)
 		free(logmsg);
 	if (err)
@@ -2326,7 +2339,7 @@ check_path_is_commitable(const char *path,
 	struct got_pathlist_entry *cpe = NULL;
 	size_t path_len = strlen(path);
 
-	TAILQ_FOREACH(cpe, commitable_paths, entry) {
+	RB_FOREACH(cpe, got_pathlist_head, commitable_paths) {
 		struct got_commitable *ct = cpe->data;
 		const char *ct_path = ct->path;
 
@@ -2364,7 +2377,7 @@ check_non_staged_files(struct got_fileindex *fileindex,
 	struct got_pathlist_entry *pe;
 	struct got_fileindex_entry *ie;
 
-	TAILQ_FOREACH(pe, paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head, paths) {
 		if (pe->path[0] == '\0')
 			continue;
 		ie = got_fileindex_entry_get(fileindex, pe->path, pe->path_len);
@@ -2436,7 +2449,7 @@ send_progress(void *arg, int ncolored, int nfound, int ntrees,
 
 		if (success) {
 			struct got_pathlist_entry *pe;
-			TAILQ_FOREACH(pe, a->delete_branches, entry) {
+			RB_FOREACH(pe, got_pathlist_head, a->delete_branches) {
 				const char *branchname = pe->path;
 				if (got_path_cmp(branchname, refname,
 				    strlen(branchname), strlen(refname)) == 0) {
@@ -2753,9 +2766,10 @@ done:
 
 static const struct got_error *
 fetch_updated_remote(const char *proto, const char *host, const char *port,
-    const char *server_path, int verbosity,
-    const struct got_remote_repo *remote, struct got_repository *repo,
-    struct got_reference *head_ref, const char *head_refname)
+    const char *server_path, const char *jumphost, const char *identity_file,
+    int verbosity, const struct got_remote_repo *remote,
+    struct got_repository *repo, struct got_reference *head_ref,
+    const char *head_refname)
 {
 	const struct got_error *err = NULL, *unlock_err = NULL;
 	struct got_pathlist_entry *pe;
@@ -2768,10 +2782,10 @@ fetch_updated_remote(const char *proto, const char *host, const char *port,
 	int fetchfd = -1;
 	pid_t fetchpid = -1;
 
-	TAILQ_INIT(&learned_refs);
-	TAILQ_INIT(&symrefs);
-	TAILQ_INIT(&wanted_branches);
-	TAILQ_INIT(&wanted_refs);
+	RB_INIT(&learned_refs);
+	RB_INIT(&symrefs);
+	RB_INIT(&wanted_branches);
+	RB_INIT(&wanted_refs);
 
 	err = got_pathlist_insert(NULL, &wanted_branches, head_refname,
 	    NULL);
@@ -2779,7 +2793,7 @@ fetch_updated_remote(const char *proto, const char *host, const char *port,
 		goto done;
 
 	err = got_fetch_connect(&fetchpid, &fetchfd, proto, host,
-	    port, server_path, verbosity);
+	    port, server_path, jumphost, identity_file, verbosity);
 	if (err)
 		goto done;
 
@@ -2796,7 +2810,7 @@ fetch_updated_remote(const char *proto, const char *host, const char *port,
 		goto done;
 
 	/* Update references provided with the pack file. */
-	TAILQ_FOREACH(pe, &learned_refs, entry) {
+	RB_FOREACH(pe, got_pathlist_head, &learned_refs) {
 		const char *refname = pe->path;
 		struct got_object_id *id = pe->data;
 		struct got_reference *ref;
@@ -2820,7 +2834,7 @@ fetch_updated_remote(const char *proto, const char *host, const char *port,
 	}
 
 	/* Set the HEAD reference if the server provided one. */
-	TAILQ_FOREACH(pe, &symrefs, entry) {
+	RB_FOREACH(pe, got_pathlist_head, &symrefs) {
 		struct got_reference *target_ref;
 		const char *refname = pe->path;
 		const char *target = pe->data;
@@ -2893,13 +2907,13 @@ done:
 const struct got_error *
 got_worktree_cvg_commit(struct got_object_id **new_commit_id,
     struct got_worktree *worktree, struct got_pathlist_head *paths,
-    const char *author, const char *committer, int allow_bad_symlinks,
-    int show_diff, int commit_conflicts,
+    const char *author, time_t author_time, const char *committer,
+    int allow_bad_symlinks, int show_diff, int commit_conflicts,
     got_worktree_commit_msg_cb commit_msg_cb, void *commit_arg,
     got_worktree_status_cb status_cb, void *status_arg,
     const char *proto, const char *host, const char *port,
-    const char *server_path, int verbosity,
-    const struct got_remote_repo *remote,
+    const char *server_path, const char *jumphost, const char *identity_file,
+    int verbosity, const struct got_remote_repo *remote,
     got_cancel_cb check_cancelled,
     struct got_repository *repo)
 {
@@ -2927,10 +2941,10 @@ got_worktree_cvg_commit(struct got_object_id **new_commit_id,
 	*new_commit_id = NULL;
 
 	memset(&cc_arg, 0, sizeof(cc_arg));
-	TAILQ_INIT(&commitable_paths);
-	TAILQ_INIT(&commit_reflist);
-	TAILQ_INIT(&tag_names);
-	TAILQ_INIT(&delete_branches);
+	RB_INIT(&commitable_paths);
+	RB_INIT(&commit_reflist);
+	RB_INIT(&tag_names);
+	RB_INIT(&delete_branches);
 
 	err = lock_worktree(worktree, LOCK_EX);
 	if (err)
@@ -2956,7 +2970,8 @@ got_worktree_cvg_commit(struct got_object_id **new_commit_id,
 	if (err)
 		goto done;
 
-	err = open_fileindex(&fileindex, &fileindex_path, worktree);
+	err = open_fileindex(&fileindex, &fileindex_path, worktree,
+	    got_repo_get_object_format(repo));
 	if (err)
 		goto done;
 
@@ -2995,7 +3010,7 @@ got_worktree_cvg_commit(struct got_object_id **new_commit_id,
 		}
 	}
 
-	TAILQ_FOREACH(pe, paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head, paths) {
 		err = worktree_status(worktree, pe->path, fileindex, repo,
 		    collect_commitables, &cc_arg, NULL, NULL, 0, 0);
 		if (err)
@@ -3009,18 +3024,18 @@ got_worktree_cvg_commit(struct got_object_id **new_commit_id,
 		}
 	}
 
-	if (TAILQ_EMPTY(&commitable_paths)) {
+	if (RB_EMPTY(&commitable_paths)) {
 		err = got_error(GOT_ERR_COMMIT_NO_CHANGES);
 		goto done;
 	}
 
-	TAILQ_FOREACH(pe, paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head, paths) {
 		err = check_path_is_commitable(pe->path, &commitable_paths);
 		if (err)
 			goto done;
 	}
 
-	TAILQ_FOREACH(pe, &commitable_paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head, &commitable_paths) {
 		struct got_commitable *ct = pe->data;
 		const char *ct_path = ct->in_repo_path;
 
@@ -3034,7 +3049,7 @@ got_worktree_cvg_commit(struct got_object_id **new_commit_id,
 	}
 
 	err = commit_worktree(new_commit_id, &commitable_paths,
-	    head_commit_id, NULL, worktree, author, committer,
+	    head_commit_id, NULL, worktree, author, author_time, committer,
 	    (diff_path && cc_arg.diff_header_shown) ? diff_path : NULL,
 	    commit_msg_cb, commit_arg, status_cb, status_arg, repo);
 	if (err)
@@ -3055,7 +3070,7 @@ got_worktree_cvg_commit(struct got_object_id **new_commit_id,
 		goto done;
 	}
 
-	err = got_pathlist_append(&commit_reflist, commit_refname,
+	err = got_pathlist_insert(&pe, &commit_reflist, commit_refname,
 	    head_refname);
 	if (err)
 		goto done;
@@ -3077,7 +3092,7 @@ got_worktree_cvg_commit(struct got_object_id **new_commit_id,
 
 	/* Attempt send to remote branch. */
 	err = got_send_connect(&sendpid, &sendfd, proto, host, port,
-	    server_path, verbosity);
+	    server_path, jumphost, identity_file, verbosity);
 	if (err)
 		goto done;
 
@@ -3098,7 +3113,8 @@ got_worktree_cvg_commit(struct got_object_id **new_commit_id,
 		 * No trivial-rebase yet; require update to be run manually.
 		 */
 		err = fetch_updated_remote(proto, host, port, server_path,
-		    verbosity, remote, repo, head_ref, head_refname);
+		    jumphost, identity_file, verbosity, remote, repo,
+		    head_ref, head_refname);
 		if (err == NULL)
 			goto done;
 		err = got_error(GOT_ERR_COMMIT_OUT_OF_DATE);
@@ -3145,7 +3161,7 @@ done:
 	unlockerr = lock_worktree(worktree, LOCK_SH);
 	if (unlockerr && err == NULL)
 		err = unlockerr;
-	TAILQ_FOREACH(pe, &commitable_paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head, &commitable_paths) {
 		struct got_commitable *ct = pe->data;
 
 		free_commitable(ct);
